"""LLM provider implementations."""

import json
from abc import ABC, abstractmethod
from typing import Any, AsyncIterator, Optional

import httpx

from sigma.core.config import LLMProvider, get_settings
from sigma.core.models import Message, MessageRole, ToolCall


class BaseLLM(ABC):
    """Base LLM provider."""
    
    def __init__(self, model: Optional[str] = None):
        self.settings = get_settings()
        self.model = model or self.settings.get_model(self.provider)
    
    @property
    @abstractmethod
    def provider(self) -> LLMProvider:
        """Provider type."""
        pass
    
    @abstractmethod
    async def generate(
        self,
        messages: list[Message],
        tools: Optional[list[dict]] = None,
        temperature: Optional[float] = None,
    ) -> tuple[str, list[ToolCall]]:
        """Generate response."""
        pass
    
    @abstractmethod
    async def stream(
        self,
        messages: list[Message],
        tools: Optional[list[dict]] = None,
        temperature: Optional[float] = None,
    ) -> AsyncIterator[str]:
        """Stream response."""
        pass


class OpenAILLM(BaseLLM):
    """OpenAI provider."""
    
    @property
    def provider(self) -> LLMProvider:
        return LLMProvider.OPENAI
    
    def _convert_messages(self, messages: list[Message]) -> list[dict]:
        """Convert to OpenAI format."""
        result = []
        for msg in messages:
            m: dict[str, Any] = {"role": msg.role.value, "content": msg.content}
            if msg.tool_calls:
                m["tool_calls"] = [
                    {
                        "id": tc.id,
                        "type": "function",
                        "function": {"name": tc.name, "arguments": json.dumps(tc.arguments)}
                    }
                    for tc in msg.tool_calls
                ]
            if msg.tool_call_id:
                m["tool_call_id"] = msg.tool_call_id
            if msg.name:
                m["name"] = msg.name
            result.append(m)
        return result
    
    async def generate(
        self,
        messages: list[Message],
        tools: Optional[list[dict]] = None,
        temperature: Optional[float] = None,
    ) -> tuple[str, list[ToolCall]]:
        api_key = self.settings.get_api_key(LLMProvider.OPENAI)
        if not api_key:
            raise ValueError("OpenAI API key not set")
        
        headers = {"Authorization": f"Bearer {api_key}", "Content-Type": "application/json"}
        data: dict[str, Any] = {
            "model": self.model,
            "messages": self._convert_messages(messages),
            "temperature": temperature or self.settings.temperature,
            "max_tokens": self.settings.max_tokens,
        }
        
        if tools:
            data["tools"] = [{"type": "function", "function": t} for t in tools]
            data["tool_choice"] = "auto"
        
        async with httpx.AsyncClient(timeout=120.0) as client:
            resp = await client.post(
                "https://api.openai.com/v1/chat/completions",
                headers=headers,
                json=data,
            )
            resp.raise_for_status()
            result = resp.json()
        
        choice = result["choices"][0]
        msg = choice["message"]
        content = msg.get("content", "") or ""
        
        tool_calls = []
        if "tool_calls" in msg and msg["tool_calls"]:
            for tc in msg["tool_calls"]:
                tool_calls.append(ToolCall(
                    id=tc["id"],
                    name=tc["function"]["name"],
                    arguments=json.loads(tc["function"]["arguments"]),
                ))
        
        return content, tool_calls
    
    async def stream(
        self,
        messages: list[Message],
        tools: Optional[list[dict]] = None,
        temperature: Optional[float] = None,
    ) -> AsyncIterator[str]:
        api_key = self.settings.get_api_key(LLMProvider.OPENAI)
        if not api_key:
            raise ValueError("OpenAI API key not set")
        
        headers = {"Authorization": f"Bearer {api_key}", "Content-Type": "application/json"}
        data: dict[str, Any] = {
            "model": self.model,
            "messages": self._convert_messages(messages),
            "temperature": temperature or self.settings.temperature,
            "max_tokens": self.settings.max_tokens,
            "stream": True,
        }
        
        async with httpx.AsyncClient(timeout=120.0) as client:
            async with client.stream(
                "POST",
                "https://api.openai.com/v1/chat/completions",
                headers=headers,
                json=data,
            ) as resp:
                async for line in resp.aiter_lines():
                    if line.startswith("data: "):
                        payload = line[6:]
                        if payload == "[DONE]":
                            break
                        chunk = json.loads(payload)
                        delta = chunk["choices"][0].get("delta", {})
                        if "content" in delta and delta["content"]:
                            yield delta["content"]


class AnthropicLLM(BaseLLM):
    """Anthropic provider."""
    
    @property
    def provider(self) -> LLMProvider:
        return LLMProvider.ANTHROPIC
    
    def _convert_messages(self, messages: list[Message]) -> tuple[Optional[str], list[dict]]:
        """Convert to Anthropic format."""
        system = None
        result = []
        
        for msg in messages:
            if msg.role == MessageRole.SYSTEM:
                system = msg.content
                continue
            
            if msg.role == MessageRole.TOOL:
                result.append({
                    "role": "user",
                    "content": [{
                        "type": "tool_result",
                        "tool_use_id": msg.tool_call_id,
                        "content": msg.content,
                    }]
                })
            elif msg.tool_calls:
                content: list[dict[str, Any]] = []
                if msg.content:
                    content.append({"type": "text", "text": msg.content})
                for tc in msg.tool_calls:
                    content.append({
                        "type": "tool_use",
                        "id": tc.id,
                        "name": tc.name,
                        "input": tc.arguments,
                    })
                result.append({"role": "assistant", "content": content})
            else:
                result.append({"role": msg.role.value, "content": msg.content})
        
        return system, result
    
    async def generate(
        self,
        messages: list[Message],
        tools: Optional[list[dict]] = None,
        temperature: Optional[float] = None,
    ) -> tuple[str, list[ToolCall]]:
        api_key = self.settings.get_api_key(LLMProvider.ANTHROPIC)
        if not api_key:
            raise ValueError("Anthropic API key not set")
        
        headers = {
            "x-api-key": api_key,
            "Content-Type": "application/json",
            "anthropic-version": "2023-06-01",
        }
        
        system, msgs = self._convert_messages(messages)
        data: dict[str, Any] = {
            "model": self.model,
            "messages": msgs,
            "max_tokens": self.settings.max_tokens,
            "temperature": temperature or self.settings.temperature,
        }
        
        if system:
            data["system"] = system
        
        if tools:
            data["tools"] = [
                {
                    "name": t["name"],
                    "description": t.get("description", ""),
                    "input_schema": t.get("parameters", {}),
                }
                for t in tools
            ]
        
        async with httpx.AsyncClient(timeout=120.0) as client:
            resp = await client.post(
                "https://api.anthropic.com/v1/messages",
                headers=headers,
                json=data,
            )
            resp.raise_for_status()
            result = resp.json()
        
        content = ""
        tool_calls = []
        
        for block in result.get("content", []):
            if block["type"] == "text":
                content += block["text"]
            elif block["type"] == "tool_use":
                tool_calls.append(ToolCall(
                    id=block["id"],
                    name=block["name"],
                    arguments=block["input"],
                ))
        
        return content, tool_calls
    
    async def stream(
        self,
        messages: list[Message],
        tools: Optional[list[dict]] = None,
        temperature: Optional[float] = None,
    ) -> AsyncIterator[str]:
        api_key = self.settings.get_api_key(LLMProvider.ANTHROPIC)
        if not api_key:
            raise ValueError("Anthropic API key not set")
        
        headers = {
            "x-api-key": api_key,
            "Content-Type": "application/json",
            "anthropic-version": "2023-06-01",
        }
        
        system, msgs = self._convert_messages(messages)
        data: dict[str, Any] = {
            "model": self.model,
            "messages": msgs,
            "max_tokens": self.settings.max_tokens,
            "temperature": temperature or self.settings.temperature,
            "stream": True,
        }
        
        if system:
            data["system"] = system
        
        async with httpx.AsyncClient(timeout=120.0) as client:
            async with client.stream(
                "POST",
                "https://api.anthropic.com/v1/messages",
                headers=headers,
                json=data,
            ) as resp:
                async for line in resp.aiter_lines():
                    if line.startswith("data: "):
                        event = json.loads(line[6:])
                        if event["type"] == "content_block_delta":
                            delta = event.get("delta", {})
                            if delta.get("type") == "text_delta":
                                yield delta.get("text", "")


class GoogleLLM(BaseLLM):
    """Google Gemini provider using REST API."""
    
    @property
    def provider(self) -> LLMProvider:
        return LLMProvider.GOOGLE
    
    def _convert_messages(self, messages: list[Message]) -> tuple[Optional[str], list[dict]]:
        """Convert to Gemini format."""
        system = None
        contents = []
        
        for msg in messages:
            if msg.role == MessageRole.SYSTEM:
                system = msg.content
                continue
            
            role = "model" if msg.role == MessageRole.ASSISTANT else "user"
            
            if msg.role == MessageRole.TOOL:
                contents.append({
                    "role": "user",
                    "parts": [{
                        "functionResponse": {
                            "name": msg.name or "tool",
                            "response": {"result": msg.content}
                        }
                    }]
                })
            elif msg.tool_calls:
                parts: list[dict[str, Any]] = []
                if msg.content:
                    parts.append({"text": msg.content})
                for tc in msg.tool_calls:
                    parts.append({
                        "functionCall": {
                            "name": tc.name,
                            "args": tc.arguments
                        }
                    })
                contents.append({"role": role, "parts": parts})
            else:
                contents.append({
                    "role": role,
                    "parts": [{"text": msg.content}]
                })
        
        return system, contents
    
    def _convert_tools(self, tools: list[dict]) -> list[dict]:
        """Convert tools to Gemini format."""
        declarations = []
        for t in tools:
            decl: dict[str, Any] = {
                "name": t["name"],
                "description": t.get("description", ""),
            }
            if "parameters" in t and t["parameters"]:
                params = t["parameters"].copy()
                # Gemini doesn't want 'additionalProperties'
                params.pop("additionalProperties", None)
                decl["parameters"] = params
            declarations.append(decl)
        return [{"functionDeclarations": declarations}]
    
    async def generate(
        self,
        messages: list[Message],
        tools: Optional[list[dict]] = None,
        temperature: Optional[float] = None,
    ) -> tuple[str, list[ToolCall]]:
        api_key = self.settings.get_api_key(LLMProvider.GOOGLE)
        if not api_key:
            raise ValueError("Google API key not set")
        
        system, contents = self._convert_messages(messages)
        
        data: dict[str, Any] = {
            "contents": contents,
            "generationConfig": {
                "temperature": temperature or self.settings.temperature,
                "maxOutputTokens": self.settings.max_tokens,
            }
        }
        
        if system:
            data["systemInstruction"] = {"parts": [{"text": system}]}
        
        if tools:
            data["tools"] = self._convert_tools(tools)
        
        url = f"https://generativelanguage.googleapis.com/v1beta/models/{self.model}:generateContent?key={api_key}"
        
        async with httpx.AsyncClient(timeout=120.0) as client:
            resp = await client.post(url, json=data)
            resp.raise_for_status()
            result = resp.json()
        
        content = ""
        tool_calls = []
        
        candidates = result.get("candidates", [])
        if candidates:
            parts = candidates[0].get("content", {}).get("parts", [])
            for i, part in enumerate(parts):
                if "text" in part:
                    content += part["text"]
                elif "functionCall" in part:
                    fc = part["functionCall"]
                    tool_calls.append(ToolCall(
                        id=f"call_{i}",
                        name=fc["name"],
                        arguments=fc.get("args", {}),
                    ))
        
        return content, tool_calls
    
    async def stream(
        self,
        messages: list[Message],
        tools: Optional[list[dict]] = None,
        temperature: Optional[float] = None,
    ) -> AsyncIterator[str]:
        api_key = self.settings.get_api_key(LLMProvider.GOOGLE)
        if not api_key:
            raise ValueError("Google API key not set")
        
        system, contents = self._convert_messages(messages)
        
        data: dict[str, Any] = {
            "contents": contents,
            "generationConfig": {
                "temperature": temperature or self.settings.temperature,
                "maxOutputTokens": self.settings.max_tokens,
            }
        }
        
        if system:
            data["systemInstruction"] = {"parts": [{"text": system}]}
        
        url = f"https://generativelanguage.googleapis.com/v1beta/models/{self.model}:streamGenerateContent?key={api_key}&alt=sse"
        
        async with httpx.AsyncClient(timeout=120.0) as client:
            async with client.stream("POST", url, json=data) as resp:
                async for line in resp.aiter_lines():
                    if line.startswith("data: "):
                        chunk = json.loads(line[6:])
                        candidates = chunk.get("candidates", [])
                        if candidates:
                            parts = candidates[0].get("content", {}).get("parts", [])
                            for part in parts:
                                if "text" in part:
                                    yield part["text"]


class OllamaLLM(BaseLLM):
    """Ollama provider."""
    
    @property
    def provider(self) -> LLMProvider:
        return LLMProvider.OLLAMA
    
    def _convert_messages(self, messages: list[Message]) -> list[dict]:
        """Convert to Ollama format."""
        result = []
        for msg in messages:
            m: dict[str, Any] = {"role": msg.role.value, "content": msg.content}
            if msg.tool_calls:
                m["tool_calls"] = [
                    {
                        "id": tc.id,
                        "type": "function",
                        "function": {"name": tc.name, "arguments": tc.arguments}
                    }
                    for tc in msg.tool_calls
                ]
            result.append(m)
        return result
    
    async def generate(
        self,
        messages: list[Message],
        tools: Optional[list[dict]] = None,
        temperature: Optional[float] = None,
    ) -> tuple[str, list[ToolCall]]:
        data: dict[str, Any] = {
            "model": self.model,
            "messages": self._convert_messages(messages),
            "stream": False,
            "options": {
                "temperature": temperature or self.settings.temperature,
                "num_predict": self.settings.max_tokens,
            }
        }
        
        if tools:
            data["tools"] = [{"type": "function", "function": t} for t in tools]
        
        async with httpx.AsyncClient(timeout=300.0) as client:
            resp = await client.post(
                f"{self.settings.ollama_base_url}/api/chat",
                json=data,
            )
            resp.raise_for_status()
            result = resp.json()
        
        msg = result.get("message", {})
        content = msg.get("content", "") or ""
        
        tool_calls = []
        if "tool_calls" in msg and msg["tool_calls"]:
            for i, tc in enumerate(msg["tool_calls"]):
                fn = tc.get("function", {})
                tool_calls.append(ToolCall(
                    id=f"call_{i}",
                    name=fn.get("name", ""),
                    arguments=fn.get("arguments", {}),
                ))
        
        return content, tool_calls
    
    async def stream(
        self,
        messages: list[Message],
        tools: Optional[list[dict]] = None,
        temperature: Optional[float] = None,
    ) -> AsyncIterator[str]:
        data: dict[str, Any] = {
            "model": self.model,
            "messages": self._convert_messages(messages),
            "stream": True,
            "options": {
                "temperature": temperature or self.settings.temperature,
                "num_predict": self.settings.max_tokens,
            }
        }
        
        async with httpx.AsyncClient(timeout=300.0) as client:
            async with client.stream(
                "POST",
                f"{self.settings.ollama_base_url}/api/chat",
                json=data,
            ) as resp:
                async for line in resp.aiter_lines():
                    if line:
                        chunk = json.loads(line)
                        msg = chunk.get("message", {})
                        if "content" in msg:
                            yield msg["content"]


class GroqLLM(BaseLLM):
    """Groq provider."""
    
    @property
    def provider(self) -> LLMProvider:
        return LLMProvider.GROQ
    
    def _convert_messages(self, messages: list[Message]) -> list[dict]:
        """Convert to Groq format (OpenAI compatible)."""
        result = []
        for msg in messages:
            m: dict[str, Any] = {"role": msg.role.value, "content": msg.content}
            if msg.tool_calls:
                m["tool_calls"] = [
                    {
                        "id": tc.id,
                        "type": "function",
                        "function": {"name": tc.name, "arguments": json.dumps(tc.arguments)}
                    }
                    for tc in msg.tool_calls
                ]
            if msg.tool_call_id:
                m["tool_call_id"] = msg.tool_call_id
            if msg.name:
                m["name"] = msg.name
            result.append(m)
        return result
    
    async def generate(
        self,
        messages: list[Message],
        tools: Optional[list[dict]] = None,
        temperature: Optional[float] = None,
    ) -> tuple[str, list[ToolCall]]:
        api_key = self.settings.get_api_key(LLMProvider.GROQ)
        if not api_key:
            raise ValueError("Groq API key not set")
        
        headers = {"Authorization": f"Bearer {api_key}", "Content-Type": "application/json"}
        data: dict[str, Any] = {
            "model": self.model,
            "messages": self._convert_messages(messages),
            "temperature": temperature or self.settings.temperature,
            "max_tokens": self.settings.max_tokens,
        }
        
        if tools:
            data["tools"] = [{"type": "function", "function": t} for t in tools]
            data["tool_choice"] = "auto"
        
        async with httpx.AsyncClient(timeout=120.0) as client:
            resp = await client.post(
                "https://api.groq.com/openai/v1/chat/completions",
                headers=headers,
                json=data,
            )
            resp.raise_for_status()
            result = resp.json()
        
        choice = result["choices"][0]
        msg = choice["message"]
        content = msg.get("content", "") or ""
        
        tool_calls = []
        if "tool_calls" in msg and msg["tool_calls"]:
            for tc in msg["tool_calls"]:
                tool_calls.append(ToolCall(
                    id=tc["id"],
                    name=tc["function"]["name"],
                    arguments=json.loads(tc["function"]["arguments"]),
                ))
        
        return content, tool_calls
    
    async def stream(
        self,
        messages: list[Message],
        tools: Optional[list[dict]] = None,
        temperature: Optional[float] = None,
    ) -> AsyncIterator[str]:
        api_key = self.settings.get_api_key(LLMProvider.GROQ)
        if not api_key:
            raise ValueError("Groq API key not set")
        
        headers = {"Authorization": f"Bearer {api_key}", "Content-Type": "application/json"}
        data: dict[str, Any] = {
            "model": self.model,
            "messages": self._convert_messages(messages),
            "temperature": temperature or self.settings.temperature,
            "max_tokens": self.settings.max_tokens,
            "stream": True,
        }
        
        async with httpx.AsyncClient(timeout=120.0) as client:
            async with client.stream(
                "POST",
                "https://api.groq.com/openai/v1/chat/completions",
                headers=headers,
                json=data,
            ) as resp:
                async for line in resp.aiter_lines():
                    if line.startswith("data: "):
                        payload = line[6:]
                        if payload == "[DONE]":
                            break
                        chunk = json.loads(payload)
                        delta = chunk["choices"][0].get("delta", {})
                        if "content" in delta and delta["content"]:
                            yield delta["content"]


class XaiLLM(BaseLLM):
    """xAI Grok provider."""
    
    @property
    def provider(self) -> LLMProvider:
        return LLMProvider.XAI
    
    def _convert_messages(self, messages: list[Message]) -> list[dict]:
        """Convert to xAI format (OpenAI compatible)."""
        result = []
        for msg in messages:
            m: dict[str, Any] = {"role": msg.role.value, "content": msg.content}
            if msg.tool_calls:
                m["tool_calls"] = [
                    {
                        "id": tc.id,
                        "type": "function",
                        "function": {"name": tc.name, "arguments": json.dumps(tc.arguments)}
                    }
                    for tc in msg.tool_calls
                ]
            if msg.tool_call_id:
                m["tool_call_id"] = msg.tool_call_id
            result.append(m)
        return result
    
    async def generate(
        self,
        messages: list[Message],
        tools: Optional[list[dict]] = None,
        temperature: Optional[float] = None,
    ) -> tuple[str, list[ToolCall]]:
        api_key = self.settings.get_api_key(LLMProvider.XAI)
        if not api_key:
            raise ValueError("xAI API key not set")
        
        headers = {"Authorization": f"Bearer {api_key}", "Content-Type": "application/json"}
        data: dict[str, Any] = {
            "model": self.model,
            "messages": self._convert_messages(messages),
            "temperature": temperature or self.settings.temperature,
            "max_tokens": self.settings.max_tokens,
        }
        
        if tools:
            data["tools"] = [{"type": "function", "function": t} for t in tools]
        
        async with httpx.AsyncClient(timeout=120.0) as client:
            resp = await client.post(
                "https://api.x.ai/v1/chat/completions",
                headers=headers,
                json=data,
            )
            resp.raise_for_status()
            result = resp.json()
        
        choice = result["choices"][0]
        msg = choice["message"]
        content = msg.get("content", "") or ""
        
        tool_calls = []
        if "tool_calls" in msg and msg["tool_calls"]:
            for tc in msg["tool_calls"]:
                tool_calls.append(ToolCall(
                    id=tc["id"],
                    name=tc["function"]["name"],
                    arguments=json.loads(tc["function"]["arguments"]),
                ))
        
        return content, tool_calls
    
    async def stream(
        self,
        messages: list[Message],
        tools: Optional[list[dict]] = None,
        temperature: Optional[float] = None,
    ) -> AsyncIterator[str]:
        api_key = self.settings.get_api_key(LLMProvider.XAI)
        if not api_key:
            raise ValueError("xAI API key not set")
        
        headers = {"Authorization": f"Bearer {api_key}", "Content-Type": "application/json"}
        data: dict[str, Any] = {
            "model": self.model,
            "messages": self._convert_messages(messages),
            "temperature": temperature or self.settings.temperature,
            "max_tokens": self.settings.max_tokens,
            "stream": True,
        }
        
        async with httpx.AsyncClient(timeout=120.0) as client:
            async with client.stream(
                "POST",
                "https://api.x.ai/v1/chat/completions",
                headers=headers,
                json=data,
            ) as resp:
                async for line in resp.aiter_lines():
                    if line.startswith("data: "):
                        payload = line[6:]
                        if payload == "[DONE]":
                            break
                        chunk = json.loads(payload)
                        delta = chunk["choices"][0].get("delta", {})
                        if "content" in delta and delta["content"]:
                            yield delta["content"]


def get_llm(provider: Optional[LLMProvider] = None, model: Optional[str] = None) -> BaseLLM:
    """Get LLM instance."""
    settings = get_settings()
    provider = provider or settings.default_provider
    
    providers = {
        LLMProvider.OPENAI: OpenAILLM,
        LLMProvider.ANTHROPIC: AnthropicLLM,
        LLMProvider.GOOGLE: GoogleLLM,
        LLMProvider.OLLAMA: OllamaLLM,
        LLMProvider.GROQ: GroqLLM,
        LLMProvider.XAI: XaiLLM,
    }
    
    cls = providers.get(provider)
    if not cls:
        raise ValueError(f"Unknown provider: {provider}")
    
    return cls(model=model)



"""Sigma CLI - Professional financial research interface."""

import asyncio
import sys
import re
from typing import Any, Optional
import time

from rich.console import Console
from rich.panel import Panel
from rich.text import Text
from rich.table import Table
from rich.markdown import Markdown
from rich.padding import Padding
from rich.box import ROUNDED, SIMPLE, HEAVY, DOUBLE
from rich.columns import Columns
from rich import box

from sigma.core.agent import SigmaAgent
from sigma.core.config import LLMProvider, get_settings


console = Console()

# Version
VERSION = "2.0.2"

# ASCII Art Banner
BANNER = """[bold bright_blue]
 ███████╗██╗ ██████╗ ███╗   ███╗ █████╗ 
 ██╔════╝██║██╔════╝ ████╗ ████║██╔══██╗
 ███████╗██║██║  ███╗██╔████╔██║███████║
 ╚════██║██║██║   ██║██║╚██╔╝██║██╔══██║
 ███████║██║╚██████╔╝██║ ╚═╝ ██║██║  ██║
 ╚══════╝╚═╝ ╚═════╝ ╚═╝     ╚═╝╚═╝  ╚═╝[/bold bright_blue]
"""

SUB_BANNER = """[dim]━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━[/dim]
[bold white]        Institutional-Grade Financial Research Agent[/bold white]
[dim]━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━[/dim]"""


def print_banner(model: str):
    """Print the startup banner."""
    console.print()
    console.print(BANNER)
    console.print(SUB_BANNER)
    console.print()
    
    # Status line
    console.print(f"  [dim]Version:[/dim] [bright_cyan]{VERSION}[/bright_cyan]    [dim]Model:[/dim] [bright_cyan]{model}[/bright_cyan]")
    console.print()
    console.print("  [dim]Type[/dim] [bold bright_yellow]/help[/bold bright_yellow] [dim]for commands[/dim]    [dim]Type[/dim] [bold bright_yellow]/quit[/bold bright_yellow] [dim]to exit[/dim]")
    console.print()


def print_help():
    """Print comprehensive help."""
    console.print()
    
    # Commands table
    commands = Table(title="[bold bright_cyan]Commands[/bold bright_cyan]", box=ROUNDED, show_header=True, header_style="bold")
    commands.add_column("Command", style="bright_yellow")
    commands.add_column("Description", style="white")
    
    commands.add_row("/help, /h, /?", "Show this help message")
    commands.add_row("/model <provider>", "Switch LLM provider (openai, anthropic, google, groq, ollama)")
    commands.add_row("/mode <mode>", "Switch analysis mode (default, technical, fundamental, quant)")
    commands.add_row("/clear", "Clear conversation history")
    commands.add_row("/status", "Show current configuration")
    commands.add_row("/chart <symbol>", "Quick price chart for a symbol")
    commands.add_row("/compare <sym1> <sym2>...", "Quick comparison chart")
    commands.add_row("/backtest <symbol> <strategy>", "Generate backtest algorithm")
    commands.add_row("/lean setup", "Setup LEAN Engine for backtesting")
    commands.add_row("/lean run <symbol> <strategy>", "Run backtest with LEAN Engine")
    commands.add_row("/quit, /exit, /q", "Exit Sigma")
    
    console.print(commands)
    console.print()
    
    # Analysis modes
    modes = Table(title="[bold bright_cyan]Analysis Modes[/bold bright_cyan]", box=ROUNDED, show_header=True, header_style="bold")
    modes.add_column("Mode", style="bright_yellow")
    modes.add_column("Description", style="white")
    
    modes.add_row("default", "Comprehensive analysis using all available tools")
    modes.add_row("technical", "Focus on technical indicators, charts, and price action")
    modes.add_row("fundamental", "Focus on financial statements, ratios, and valuations")
    modes.add_row("quant", "Quantitative analysis with predictions and backtesting")
    
    console.print(modes)
    console.print()
    
    # Examples
    examples = Table(title="[bold bright_cyan]Example Queries[/bold bright_cyan]", box=ROUNDED, show_header=False)
    examples.add_column("Query", style="white")
    
    examples.add_row("[dim]── Stock Analysis ──[/dim]")
    examples.add_row("  Analyze NVDA stock")
    examples.add_row("  Is AAPL a good investment?")
    examples.add_row("  Give me a deep dive on Tesla")
    examples.add_row("")
    examples.add_row("[dim]── Charts & Visualization ──[/dim]")
    examples.add_row("  Show me a chart of MSFT for the past 6 months")
    examples.add_row("  Compare NVDA, AMD, and INTC")
    examples.add_row("  Show RSI chart for SPY")
    examples.add_row("")
    examples.add_row("[dim]── Technical Analysis ──[/dim]")
    examples.add_row("  Technical analysis on QQQ")
    examples.add_row("  What are the support levels for META?")
    examples.add_row("  Is GOOGL overbought?")
    examples.add_row("")
    examples.add_row("[dim]── Predictions & Sentiment ──[/dim]")
    examples.add_row("  Predict AMZN price for next month")
    examples.add_row("  What's the sentiment on TSLA?")
    examples.add_row("  Price forecast for Bitcoin")
    examples.add_row("")
    examples.add_row("[dim]── Backtesting & Strategy ──[/dim]")
    examples.add_row("  Create a MACD strategy backtest for AAPL")
    examples.add_row("  Generate RSI mean reversion backtest for SPY")
    examples.add_row("  List available backtest strategies")
    examples.add_row("")
    examples.add_row("[dim]── Market Overview ──[/dim]")
    examples.add_row("  What are today's market movers?")
    examples.add_row("  Show sector performance")
    examples.add_row("  How are the major indices doing?")
    
    console.print(examples)
    console.print()
    
    # Strategies
    strategies = Table(title="[bold bright_cyan]Backtest Strategies[/bold bright_cyan]", box=ROUNDED, show_header=True, header_style="bold")
    strategies.add_column("Strategy", style="bright_yellow")
    strategies.add_column("Description", style="white")
    
    strategies.add_row("sma_crossover", "Classic moving average crossover (fast/slow SMA)")
    strategies.add_row("rsi_mean_reversion", "Buy oversold, sell overbought using RSI")
    strategies.add_row("macd_momentum", "MACD histogram momentum strategy")
    strategies.add_row("bollinger_bands", "Mean reversion at Bollinger Band extremes")
    strategies.add_row("dual_momentum", "Absolute + relative momentum")
    strategies.add_row("breakout", "Donchian channel breakout strategy")
    
    console.print(strategies)
    console.print()


# Tool name mappings for cleaner display
TOOL_DISPLAY_NAMES = {
    "get_stock_quote": "Stock Quote",
    "get_stock_history": "Price History",
    "get_company_info": "Company Info",
    "get_financial_statements": "Financials",
    "get_analyst_recommendations": "Analyst Ratings",
    "get_insider_trades": "Insider Trades",
    "get_institutional_holders": "Institutions",
    "get_earnings_calendar": "Earnings",
    "get_options_chain": "Options Chain",
    "get_dividends": "Dividends",
    "compare_stocks": "Compare Stocks",
    "get_market_movers": "Market Movers",
    "get_sector_performance": "Sectors",
    "get_market_indices": "Market Indices",
    "calculate_portfolio_metrics": "Portfolio",
    "search_stocks": "Stock Search",
    "get_stock_news": "News",
    "technical_analysis": "Technicals",
    "generate_price_chart": "Price Chart",
    "generate_comparison_chart": "Comparison",
    "generate_rsi_chart": "RSI Chart",
    "generate_sector_chart": "Sectors",
    "list_backtest_strategies": "Strategies",
    "generate_backtest": "Backtest",
    "generate_custom_backtest": "Custom Backtest",
    "price_forecast": "Forecast",
    "sentiment_analysis": "Sentiment",
}

TOOL_ICONS = {
    "get_stock_quote": ">",
    "get_stock_history": ">",
    "get_company_info": ">",
    "get_financial_statements": ">",
    "get_analyst_recommendations": ">",
    "get_insider_trades": ">",
    "get_institutional_holders": ">",
    "get_earnings_calendar": ">",
    "get_options_chain": ">",
    "get_dividends": ">",
    "compare_stocks": ">",
    "get_market_movers": ">",
    "get_sector_performance": ">",
    "get_market_indices": ">",
    "calculate_portfolio_metrics": ">",
    "search_stocks": ">",
    "get_stock_news": ">",
    "technical_analysis": ">",
    "generate_price_chart": ">",
    "generate_comparison_chart": ">",
    "generate_rsi_chart": ">",
    "generate_sector_chart": ">",
    "list_backtest_strategies": ">",
    "generate_backtest": ">",
    "generate_custom_backtest": ">",
    "price_forecast": ">",
    "sentiment_analysis": ">",
}


def format_tool_description(name: str, args: dict) -> tuple[str, Optional[str]]:
    """Format tool call with description. Returns (display_name, detail)."""
    display_name = TOOL_DISPLAY_NAMES.get(name, name.replace('_', ' ').title())
    icon = TOOL_ICONS.get(name, "●")
    
    symbol = args.get('symbol', '')
    symbols = args.get('symbols', [])
    query = args.get('query', '')
    strategy = args.get('strategy', '')
    
    if symbol and strategy:
        return f"{icon} {display_name}", f"{symbol.upper()} - {strategy}"
    elif symbol:
        return f"{icon} {display_name}", symbol.upper()
    elif symbols:
        return f"{icon} {display_name}", ', '.join(s.upper() for s in symbols[:3])
    elif query:
        return f"{icon} {display_name}", query[:30]
    else:
        return f"{icon} {display_name}", None


class SigmaUI:
    """Professional terminal UI for Sigma."""
    
    def __init__(self):
        self.settings = get_settings()
        self.agent: Optional[SigmaAgent] = None
        self.provider = self.settings.default_provider
        self.tool_calls: list[tuple[str, float]] = []
        self.start_time = 0.0
        self.custom_model: Optional[str] = None
        self.mode = "default"  # default, technical, fundamental, quant
        self.chart_output: Optional[str] = None  # Store chart for display
    
    def _init_agent(self):
        """Initialize agent."""
        self.agent = SigmaAgent(provider=self.provider, model=self.custom_model)
    
    def _get_model_display(self) -> str:
        """Get model name for display."""
        if self.custom_model:
            return self.custom_model
        return self.settings.get_model(self.provider)
    
    def on_tool_start(self, name: str, args: dict):
        """Called when tool starts."""
        display_name, detail = format_tool_description(name, args)
        if detail:
            console.print(f"    [bright_cyan]{display_name}[/bright_cyan] [dim]({detail})[/dim]")
        else:
            console.print(f"    [bright_cyan]{display_name}[/bright_cyan]")
    
    def on_tool_end(self, name: str, result: Any, duration_ms: float):
        """Called when tool ends."""
        self.tool_calls.append((name, duration_ms))
        
        # Check for chart output
        if isinstance(result, dict) and result.get("display_as_chart"):
            self.chart_output = result.get("chart", "")
    
    def on_thinking(self, content: str):
        """Called when agent is thinking."""
        pass
    
    def on_response(self, content: str):
        """Called with final response."""
        pass
    
    async def process_query(self, query: str):
        """Process a query."""
        if self.agent is None:
            self._init_agent()
        
        assert self.agent is not None
        
        self.tool_calls = []
        self.start_time = time.time()
        self.chart_output = None
        
        console.print()
        console.print("  [dim]╭─ Researching...[/dim]")
        
        # Run agent
        response = await self.agent.run(
            query,
            on_tool_start=self.on_tool_start,
            on_tool_end=self.on_tool_end,
            on_thinking=self.on_thinking,
            on_response=self.on_response,
        )
        
        elapsed = time.time() - self.start_time
        
        # Print data sources summary
        if self.tool_calls:
            console.print(f"  [dim]╰─[/dim] [bright_green]Called {len(self.tool_calls)} data sources[/bright_green] [dim]in {elapsed:.1f}s[/dim]")
        
        console.print()
        
        # Print chart if generated
        if self.chart_output:
            console.print(self.chart_output)
            console.print()
        
        # Print response
        self._display_response(response)
        
        console.print()
    
    def _display_response(self, response: str):
        """Display the response with professional formatting."""
        lines = response.split('\n')
        in_table = False
        table_lines: list[str] = []
        
        for line in lines:
            # Check if this is a table line
            if '|' in line and ('---' in line or line.strip().startswith('|')):
                if not in_table:
                    in_table = True
                    table_lines = []
                table_lines.append(line)
                continue
            elif in_table and line.strip():
                if '|' in line:
                    table_lines.append(line)
                    continue
                else:
                    self._print_table(table_lines)
                    in_table = False
                    table_lines = []
            elif in_table and not line.strip():
                self._print_table(table_lines)
                in_table = False
                table_lines = []
            
            # Format and print the line
            formatted = self._format_line(line)
            if formatted is not None:
                console.print(formatted)
        
        # Print any remaining table
        if table_lines:
            self._print_table(table_lines)
    
    def _format_line(self, line: str) -> Optional[str]:
        """Format a single line with proper markdown rendering."""
        stripped = line.strip()
        
        # Empty line
        if not stripped:
            return ""
        
        # Headers
        if stripped.startswith('#'):
            header_match = re.match(r'^(#+)\s*(.+)$', stripped)
            if header_match:
                level = len(header_match.group(1))
                text = header_match.group(2)
                if level == 1:
                    return f"\n[bold bright_white]{text}[/bold bright_white]"
                elif level == 2:
                    return f"\n[bold bright_cyan]▸ {text}[/bold bright_cyan]"
                else:
                    return f"\n[bold]{text}[/bold]"
        
        # Bold-only lines
        if stripped.startswith('**') and stripped.endswith('**') and stripped.count('**') == 2:
            text = stripped[2:-2]
            return f"\n[bold bright_yellow]▸ {text}[/bold bright_yellow]"
        
        # Lines starting with bold
        bold_start = re.match(r'^\*\*(.+?)\*\*(.*)$', stripped)
        if bold_start:
            bold_part = bold_start.group(1)
            rest = bold_start.group(2)
            rest = re.sub(r'\*\*(.+?)\*\*', r'[bold]\1[/bold]', rest)
            rest = re.sub(r'`([^`]+)`', r'[bright_cyan]\1[/bright_cyan]', rest)
            return f"[bold bright_green]{bold_part}[/bold bright_green]{rest}"
        
        # Bullet points
        if stripped.startswith('- ') or stripped.startswith('* '):
            bullet_text = stripped[2:]
            bullet_text = re.sub(r'\*\*(.+?)\*\*', r'[bold]\1[/bold]', bullet_text)
            bullet_text = re.sub(r'`([^`]+)`', r'[bright_cyan]\1[/bright_cyan]', bullet_text)
            return f"  [dim]•[/dim] {bullet_text}"
        
        # Numbered lists
        numbered = re.match(r'^(\d+)\.\s+(.+)$', stripped)
        if numbered:
            num = numbered.group(1)
            text = numbered.group(2)
            text = re.sub(r'\*\*(.+?)\*\*', r'[bold]\1[/bold]', text)
            text = re.sub(r'`([^`]+)`', r'[bright_cyan]\1[/bright_cyan]', text)
            return f"  [bright_cyan]{num}.[/bright_cyan] {text}"
        
        # Regular line - handle inline formatting
        formatted = stripped
        formatted = re.sub(r'\*\*(.+?)\*\*', r'[bold]\1[/bold]', formatted)
        formatted = re.sub(r'`([^`]+)`', r'[bright_cyan]\1[/bright_cyan]', formatted)
        
        return formatted
    
    def _print_table(self, lines: list[str]):
        """Print a markdown table as a Rich table."""
        if len(lines) < 2:
            return
        
        header_line = lines[0]
        headers = [h.strip() for h in header_line.split('|') if h.strip()]
        
        table = Table(box=ROUNDED, show_header=True, header_style="bold bright_cyan")
        for h in headers:
            table.add_column(h)
        
        for line in lines[2:]:
            if '---' in line:
                continue
            cells = [c.strip() for c in line.split('|') if c.strip()]
            if cells:
                # Color code values
                colored_cells = []
                for cell in cells:
                    if cell.startswith('+') or cell.lower() in ['buy', 'bullish', 'strong buy']:
                        colored_cells.append(f"[green]{cell}[/green]")
                    elif cell.startswith('-') or cell.lower() in ['sell', 'bearish']:
                        colored_cells.append(f"[red]{cell}[/red]")
                    else:
                        colored_cells.append(cell)
                table.add_row(*colored_cells)
        
        console.print()
        console.print(Padding(table, (0, 2)))
    
    async def quick_chart(self, symbol: str, period: str = "3mo"):
        """Generate a quick chart."""
        from sigma.tools.charts import create_price_chart
        chart = create_price_chart(symbol, period)
        console.print()
        console.print(chart)
        console.print()
    
    async def quick_compare(self, symbols: list[str], period: str = "3mo"):
        """Generate a quick comparison chart."""
        from sigma.tools.charts import create_comparison_chart
        chart = create_comparison_chart(symbols, period)
        console.print()
        console.print(chart)
        console.print()
    
    def _handle_lean_command(self, args: list[str]):
        """Handle /lean commands for LEAN Engine backtesting."""
        from sigma.tools.backtest import setup_lean_engine, run_lean_backtest, check_lean_status, get_available_strategies
        
        if not args:
            # Show help menu
            strategies = get_available_strategies()
            console.print()
            console.print(Panel(
                "\n".join([
                    "[bold cyan]LEAN Engine Backtesting[/bold cyan]",
                    "",
                    "[bold]Commands:[/bold]",
                    "  /lean setup              - Setup LEAN Engine (one-time)",
                    "  /lean run <SYM> <STRAT>  - Run comprehensive backtest",
                    "  /lean status             - Check LEAN setup status",
                    "  /lean strategies         - List available strategies",
                    "",
                    "[bold]Quick Start:[/bold]",
                    "  /lean run AAPL sma_crossover",
                    "  /lean run TSLA macd_momentum",
                    "  /lean run NVDA rsi_mean_reversion",
                    "",
                    "[bold]Available Strategies:[/bold]",
                    *[f"  [cyan]{name}[/cyan] - {s['description'][:50]}..." for name, s in list(strategies.items())[:3]],
                    "  [dim]...use /lean strategies for full list[/dim]"
                ]),
                title="[bold bright_cyan]LEAN Backtest Engine[/bold bright_cyan]",
                border_style="bright_blue"
            ))
            console.print()
            return
        
        subcmd = args[0].lower()
        
        if subcmd == "help":
            self._handle_lean_command([])
            return
        
        if subcmd == "strategies":
            strategies = get_available_strategies()
            console.print()
            from rich.table import Table
            table = Table(title="Available Backtest Strategies", border_style="bright_blue")
            table.add_column("Strategy", style="cyan")
            table.add_column("Description", style="dim")
            table.add_column("Parameters", style="yellow")
            
            for name, s in strategies.items():
                params = ", ".join([f"{k}={v}" for k, v in s.get("default_params", {}).items()])
                table.add_row(name, s["description"][:60], params[:40])
            
            console.print(table)
            console.print()
            return
        
        if subcmd == "status":
            status = check_lean_status()
            console.print()
            console.print(Panel(
                "\n".join([
                    f"[dim]Docker Installed:[/dim] {'[green]Yes[/green]' if status['docker_installed'] else '[red]No[/red]'}",
                    f"[dim]Docker Running:[/dim] {'[green]Yes[/green]' if status['docker_running'] else '[red]No[/red]'}",
                    f"[dim]LEAN Image:[/dim] {'[green]Pulled[/green]' if status['lean_image_pulled'] else '[yellow]Not pulled[/yellow]'}",
                    f"[dim]Workspace:[/dim] {'[green]Ready[/green]' if status['workspace_initialized'] else '[yellow]Not initialized[/yellow]'}",
                    "",
                    *status['instructions']
                ]),
                title="[bold bright_cyan]LEAN Engine Status[/bold bright_cyan]",
                border_style="bright_blue"
            ))
            console.print()
            return
        
        if subcmd == "setup":
            console.print()
            console.print("[bright_cyan]Setting up LEAN Engine...[/bright_cyan]")
            console.print()
            
            result = setup_lean_engine()
            
            for step in result["steps_completed"]:
                console.print(f"  [green]✓[/green] {step}")
            
            for error in result.get("errors", []):
                console.print(f"  [red]✗[/red] {error}")
            
            console.print()
            if result["success"]:
                console.print(Panel(
                    "\n".join(result["next_steps"]),
                    title="[bold bright_green]Setup Complete[/bold bright_green]",
                    border_style="green"
                ))
            else:
                console.print(Panel(
                    "\n".join(result.get("next_steps", ["Setup failed. Check errors above."])),
                    title="[bold red]Setup Incomplete[/bold red]",
                    border_style="red"
                ))
            console.print()
        
        elif subcmd == "run":
            if len(args) < 3:
                console.print("\n  [red]Usage:[/red] /lean run <symbol> <strategy>\n")
                console.print("  [dim]Strategies:[/dim] sma_crossover, rsi_mean_reversion, macd_momentum, bollinger_bands, dual_momentum, breakout\n")
                return
            
            symbol = args[1].upper()
            strategy = args[2].lower()
            
            console.print()
            console.print(f"[bright_cyan]Running Comprehensive Backtest: {symbol} - {strategy}[/bright_cyan]")
            console.print()
            
            result = run_lean_backtest(symbol, strategy)
            
            for step in result["steps"]:
                console.print(f"  [dim]>[/dim] {step}")
            
            console.print()
            
            if result["status"] == "success":
                metrics = result.get("metrics", {})
                
                # Performance metrics panel
                perf = metrics.get("performance", {})
                console.print(Panel(
                    "\n".join([
                        f"  [bold cyan]Initial Capital:[/bold cyan]    {perf.get('initial_capital', 'N/A')}",
                        f"  [bold cyan]Final Equity:[/bold cyan]       {perf.get('final_equity', 'N/A')}",
                        f"  [bold cyan]Total Return:[/bold cyan]       {perf.get('total_return', 'N/A')}",
                        f"  [bold cyan]Annual Return:[/bold cyan]      {perf.get('annual_return', 'N/A')}",
                        f"  [bold cyan]Buy & Hold:[/bold cyan]         {perf.get('buy_hold_return', 'N/A')}",
                        f"  [bold cyan]Alpha:[/bold cyan]              {perf.get('alpha', 'N/A')}",
                    ]),
                    title=f"[bold bright_green]Performance - {symbol} {strategy}[/bold bright_green]",
                    border_style="green"
                ))
                
                # Risk metrics panel
                risk = metrics.get("risk", {})
                console.print(Panel(
                    "\n".join([
                        f"  [bold yellow]Max Drawdown:[/bold yellow]      {risk.get('max_drawdown', 'N/A')}",
                        f"  [bold yellow]Volatility:[/bold yellow]        {risk.get('volatility', 'N/A')}",
                        f"  [bold yellow]Sharpe Ratio:[/bold yellow]      {risk.get('sharpe_ratio', 'N/A')}",
                        f"  [bold yellow]Sortino Ratio:[/bold yellow]     {risk.get('sortino_ratio', 'N/A')}",
                        f"  [bold yellow]Calmar Ratio:[/bold yellow]      {risk.get('calmar_ratio', 'N/A')}",
                    ]),
                    title="[bold bright_yellow]Risk Metrics[/bold bright_yellow]",
                    border_style="yellow"
                ))
                
                # Trade statistics panel
                trades = metrics.get("trades", {})
                console.print(Panel(
                    "\n".join([
                        f"  [bold magenta]Total Trades:[/bold magenta]       {trades.get('total_trades', 'N/A')}",
                        f"  [bold magenta]Win Rate:[/bold magenta]           {trades.get('win_rate', 'N/A')}",
                        f"  [bold magenta]Profit Factor:[/bold magenta]      {trades.get('profit_factor', 'N/A')}",
                        f"  [bold magenta]Avg Win:[/bold magenta]            {trades.get('avg_win', 'N/A')}",
                        f"  [bold magenta]Avg Loss:[/bold magenta]           {trades.get('avg_loss', 'N/A')}",
                        f"  [bold magenta]Avg Holding:[/bold magenta]        {trades.get('avg_holding_days', 'N/A')} days",
                    ]),
                    title="[bold bright_magenta]Trade Statistics[/bold bright_magenta]",
                    border_style="magenta"
                ))
                
                # Display charts
                charts = result.get("charts", {})
                
                if charts.get("equity_curve"):
                    console.print()
                    console.print(charts["equity_curve"])
                
                if charts.get("drawdown"):
                    console.print()
                    console.print(charts["drawdown"])
                
                if charts.get("trade_pnl"):
                    console.print()
                    console.print(charts["trade_pnl"])
                
                if charts.get("monthly_returns"):
                    console.print()
                    console.print(charts["monthly_returns"])
                
                # Recent trades table
                trade_list = result.get("trades", [])
                if trade_list:
                    console.print()
                    from rich.table import Table
                    table = Table(title="Recent Trades", border_style="dim")
                    table.add_column("Date", style="dim")
                    table.add_column("Action", style="bold")
                    table.add_column("Price", justify="right")
                    table.add_column("Shares", justify="right")
                    table.add_column("P&L", justify="right")
                    
                    for trade in trade_list[-10:]:
                        action_style = "green" if trade.get("action") == "BUY" else "red"
                        pnl = trade.get("pnl")
                        pnl_str = f"${pnl:+,.2f}" if pnl else "-"
                        pnl_style = "green" if pnl and pnl > 0 else "red" if pnl else "dim"
                        table.add_row(
                            trade.get("date", ""),
                            f"[{action_style}]{trade.get('action', '')}[/{action_style}]",
                            f"${trade.get('price', 0):,.2f}",
                            str(trade.get("shares", "")),
                            f"[{pnl_style}]{pnl_str}[/{pnl_style}]"
                        )
                    console.print(table)
                
                # Monthly returns
                monthly = result.get("monthly_returns", [])
                if monthly:
                    console.print()
                    from rich.table import Table
                    mtable = Table(title="Monthly Returns", border_style="dim")
                    mtable.add_column("Month", style="dim")
                    mtable.add_column("Return", justify="right")
                    
                    for m in monthly[-12:]:
                        ret = m.get("return", 0)
                        ret_style = "green" if ret > 0 else "red"
                        mtable.add_row(m.get("month", ""), f"[{ret_style}]{ret:+.2f}%[/{ret_style}]")
                    console.print(mtable)
                
                # QuantConnect instructions
                console.print()
                console.print(Panel(
                    "\n".join(result.get("quantconnect_instructions", [])),
                    title="[bold bright_blue]QuantConnect Cloud (Institutional Data)[/bold bright_blue]",
                    border_style="blue"
                ))
                
            else:
                console.print(Panel(
                    f"[red]Error:[/red] {result.get('error', 'Unknown error')[:200]}",
                    title="[bold red]Backtest Failed[/bold red]",
                    border_style="red"
                ))
            
            # Always show the algorithm file location
            if result.get("algorithm_file"):
                console.print()
                console.print(f"  [dim]LEAN Algorithm saved:[/dim] {result['algorithm_file']}")
            console.print()
        
        elif subcmd == "status":
            self._handle_lean_command([])  # Same as no args
        
        else:
            console.print(f"\n  [red]Unknown lean command:[/red] {subcmd}")
            console.print("  [dim]Available:[/dim] /lean setup, /lean run <symbol> <strategy>, /lean status\n")

    def handle_command(self, cmd: str) -> bool:
        """Handle slash command. Returns True to continue, False to quit."""
        parts = cmd.strip().split()
        command = parts[0].lower()
        args = parts[1:] if len(parts) > 1 else []
        
        if command in ["/quit", "/exit", "/q"]:
            return False
        
        elif command in ["/help", "/h", "/?"]:
            print_help()
        
        elif command == "/model":
            if not args:
                available = self.settings.get_available_providers()
                console.print()
                console.print(f"  [dim]Current model:[/dim] [bright_cyan]{self._get_model_display()}[/bright_cyan]")
                console.print(f"  [dim]Available:[/dim] {', '.join(p.value for p in available)}")
                console.print()
            else:
                try:
                    self.provider = LLMProvider(args[0].lower())
                    self.agent = None
                    console.print(f"\n  [bright_green]✓[/bright_green] Switched to [bright_cyan]{self._get_model_display()}[/bright_cyan]\n")
                except ValueError:
                    console.print(f"\n  [red]✗[/red] Unknown provider: {args[0]}\n")
        
        elif command == "/mode":
            if not args:
                console.print()
                console.print(f"  [dim]Current mode:[/dim] [bright_cyan]{self.mode}[/bright_cyan]")
                console.print(f"  [dim]Available:[/dim] default, technical, fundamental, quant")
                console.print()
            else:
                mode = args[0].lower()
                if mode in ["default", "technical", "fundamental", "quant"]:
                    self.mode = mode
                    console.print(f"\n  [bright_green]✓[/bright_green] Switched to [bright_cyan]{mode}[/bright_cyan] mode\n")
                else:
                    console.print(f"\n  [red]✗[/red] Unknown mode: {mode}\n")
        
        elif command == "/clear":
            if self.agent:
                self.agent.clear()
            console.print("\n  [bright_green]✓[/bright_green] Conversation cleared\n")
        
        elif command == "/status":
            console.print()
            console.print(f"  [dim]Provider:[/dim] [bright_cyan]{self.provider.value}[/bright_cyan]")
            console.print(f"  [dim]Model:[/dim] [bright_cyan]{self._get_model_display()}[/bright_cyan]")
            console.print(f"  [dim]Mode:[/dim] [bright_cyan]{self.mode}[/bright_cyan]")
            available = self.settings.get_available_providers()
            console.print(f"  [dim]Available providers:[/dim] {', '.join(p.value for p in available)}")
            if self.agent:
                stats = self.agent.get_stats()
                console.print(f"  [dim]Tools called this session:[/dim] {stats['tools_called']}")
            console.print()
        
        elif command == "/chart":
            if not args:
                console.print("\n  [red]Usage:[/red] /chart <symbol> [period]\n")
            else:
                symbol = args[0].upper()
                period = args[1] if len(args) > 1 else "3mo"
                try:
                    asyncio.get_event_loop().run_until_complete(self.quick_chart(symbol, period))
                except RuntimeError:
                    asyncio.run(self.quick_chart(symbol, period))
        
        elif command == "/compare":
            if len(args) < 2:
                console.print("\n  [red]Usage:[/red] /compare <symbol1> <symbol2> ...\n")
            else:
                symbols = [s.upper() for s in args[:5]]
                try:
                    asyncio.get_event_loop().run_until_complete(self.quick_compare(symbols))
                except RuntimeError:
                    asyncio.run(self.quick_compare(symbols))
        
        elif command == "/backtest":
            if len(args) < 2:
                console.print("\n  [red]Usage:[/red] /backtest <symbol> <strategy>\n")
                console.print("  [dim]Strategies:[/dim] sma_crossover, rsi_mean_reversion, macd_momentum, bollinger_bands, dual_momentum, breakout\n")
            else:
                symbol = args[0]
                strategy = args[1]
                try:
                    loop = asyncio.get_event_loop()
                    loop.run_until_complete(
                        self.process_query(f"Generate a {strategy} backtest for {symbol}")
                    )
                except RuntimeError:
                    asyncio.run(
                        self.process_query(f"Generate a {strategy} backtest for {symbol}")
                    )
        
        elif command == "/lean":
            self._handle_lean_command(args)
        
        else:
            console.print(f"\n  [red]Unknown command:[/red] {command}. Type /help for commands.\n")
        
        return True
    
    async def run(self):
        """Run the interactive loop."""
        print_banner(self._get_model_display())
        
        while True:
            try:
                # Professional prompt
                prompt_line = f"[bold bright_yellow]σ[/bold bright_yellow] [dim]›[/dim] "
                query = console.input(prompt_line).strip()
                
                if not query:
                    continue
                
                # Handle commands
                if query.startswith("/"):
                    if not self.handle_command(query):
                        console.print("\n  [dim]Goodbye! May your trades be ever profitable.[/dim] 📈\n")
                        break
                    continue
                
                # Process query
                await self.process_query(query)
                
            except KeyboardInterrupt:
                console.print("\n")
                continue
            except EOFError:
                console.print("\n  [dim]Goodbye! May your trades be ever profitable.[/dim] 📈\n")
                break
            except Exception as e:
                console.print(f"\n  [red]Error:[/red] {str(e)}\n")


def main():
    """Main entry point."""
    import argparse
    
    parser = argparse.ArgumentParser(
        prog="sigma",
        description="Sigma - Institutional-Grade Financial Research Agent",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
Examples:
  sigma                    Start interactive mode
  sigma --setup            Run setup wizard
  sigma "Analyze AAPL"     Direct query mode
  sigma --version          Show version

Inside Sigma:
  /help                    Show all commands
  /model openai            Switch to OpenAI
  /lean run TSLA macd      Run backtest
        """
    )
    
    parser.add_argument(
        "query",
        nargs="?",
        help="Direct query to analyze (optional)"
    )
    parser.add_argument(
        "--setup",
        action="store_true",
        help="Run the setup wizard"
    )
    parser.add_argument(
        "--reset",
        action="store_true",
        help="Reset configuration and run setup"
    )
    parser.add_argument(
        "--version", "-v",
        action="version",
        version=f"Sigma {VERSION}"
    )
    parser.add_argument(
        "--model", "-m",
        choices=["openai", "anthropic", "google", "groq", "xai", "ollama"],
        help="Override default AI model"
    )
    
    args = parser.parse_args()
    
    # Handle setup
    from sigma.setup import ensure_setup, run_setup, is_setup_complete, CONFIG_DIR
    import shutil
    
    if args.reset:
        if CONFIG_DIR.exists():
            shutil.rmtree(CONFIG_DIR)
        run_setup(force=True)
    elif args.setup:
        run_setup(force=True)
    else:
        # Ensure setup is done
        ensure_setup()
    
    # Create UI with optional model override
    ui = SigmaUI()
    if args.model:
        try:
            ui.provider = LLMProvider(args.model)
            ui.agent = None  # Force agent reload
        except ValueError:
            pass
    
    # Handle direct query or interactive mode
    if args.query:
        # Direct query mode
        async def run_query():
            print_banner(ui._get_model_display())
            await ui.process_query(args.query)
        
        try:
            asyncio.run(run_query())
        except KeyboardInterrupt:
            console.print("\n")
    else:
        # Interactive mode
        try:
            asyncio.run(ui.run())
        except KeyboardInterrupt:
            console.print("\n  [dim]Goodbye! May your trades be ever profitable.[/dim]\n")


if __name__ == "__main__":
    main()



"""Configuration for Sigma."""

from enum import Enum
from pathlib import Path
from typing import Optional

from pydantic import Field, SecretStr
from pydantic_settings import BaseSettings, SettingsConfigDict


class LLMProvider(str, Enum):
    """LLM providers."""
    OPENAI = "openai"
    ANTHROPIC = "anthropic"
    GOOGLE = "google"
    OLLAMA = "ollama"
    XAI = "xai"
    GROQ = "groq"


class Settings(BaseSettings):
    """Application settings."""
    
    model_config = SettingsConfigDict(
        env_file=".env",
        env_file_encoding="utf-8",
        extra="ignore",
    )
    
    # App
    app_name: str = "Sigma"
    debug: bool = False
    
    # LLM
    default_provider: LLMProvider = LLMProvider.GOOGLE
    
    # API Keys
    openai_api_key: Optional[SecretStr] = Field(default=None, alias="OPENAI_API_KEY")
    anthropic_api_key: Optional[SecretStr] = Field(default=None, alias="ANTHROPIC_API_KEY")
    google_api_key: Optional[SecretStr] = Field(default=None, alias="GOOGLE_API_KEY")
    xai_api_key: Optional[SecretStr] = Field(default=None, alias="XAI_API_KEY")
    groq_api_key: Optional[SecretStr] = Field(default=None, alias="GROQ_API_KEY")
    
    # Models
    openai_model: str = "gpt-4o"
    anthropic_model: str = "claude-sonnet-4-20250514"
    google_model: str = "gemini-2.0-flash"
    ollama_model: str = "llama3.2"
    xai_model: str = "grok-beta"
    groq_model: str = "llama-3.3-70b-versatile"
    
    # Ollama
    ollama_base_url: str = "http://127.0.0.1:11434"
    
    # Search
    exa_api_key: Optional[SecretStr] = Field(default=None, alias="EXASEARCH_API_KEY")
    tavily_api_key: Optional[SecretStr] = Field(default=None, alias="TAVILY_API_KEY")
    serper_api_key: Optional[SecretStr] = Field(default=None, alias="SERPER_API_KEY")
    
    # Financial
    financial_datasets_api_key: Optional[SecretStr] = Field(default=None, alias="FINANCIAL_DATASETS_API_KEY")
    fmp_api_key: Optional[SecretStr] = Field(default=None, alias="FMP_API_KEY")
    polygon_api_key: Optional[SecretStr] = Field(default=None, alias="POLYGON_API_KEY")
    alpha_vantage_api_key: Optional[SecretStr] = Field(default=None, alias="ALPHA_VANTAGE_API_KEY")
    
    # Agent
    max_iterations: int = 25
    max_tokens: int = 8192
    temperature: float = 0.1
    
    # Storage
    cache_dir: Path = Path.home() / ".sigma" / "cache"
    db_path: Path = Path.home() / ".sigma" / "sigma.db"
    
    def get_api_key(self, provider: LLMProvider) -> Optional[str]:
        """Get API key for provider."""
        keys = {
            LLMProvider.OPENAI: self.openai_api_key,
            LLMProvider.ANTHROPIC: self.anthropic_api_key,
            LLMProvider.GOOGLE: self.google_api_key,
            LLMProvider.XAI: self.xai_api_key,
            LLMProvider.GROQ: self.groq_api_key,
        }
        key = keys.get(provider)
        return key.get_secret_value() if key else None
    
    def get_model(self, provider: Optional[LLMProvider] = None) -> str:
        """Get model for provider."""
        provider = provider or self.default_provider
        models = {
            LLMProvider.OPENAI: self.openai_model,
            LLMProvider.ANTHROPIC: self.anthropic_model,
            LLMProvider.GOOGLE: self.google_model,
            LLMProvider.OLLAMA: self.ollama_model,
            LLMProvider.XAI: self.xai_model,
            LLMProvider.GROQ: self.groq_model,
        }
        return models.get(provider, self.google_model)
    
    def get_available_providers(self) -> list[LLMProvider]:
        """Get providers with API keys."""
        available = []
        for p in LLMProvider:
            if p == LLMProvider.OLLAMA:
                available.append(p)
            elif self.get_api_key(p):
                available.append(p)
        return available


_settings: Optional[Settings] = None


def get_settings() -> Settings:
    """Get settings singleton."""
    global _settings
    if _settings is None:
        _settings = Settings()
    return _settings



"""Tests for Sigma financial tools."""

import pytest
from sigma.tools.financial import (
    get_stock_quote,
    get_company_info,
    get_stock_history,
    get_analyst_recommendations,
    technical_analysis,
    compare_stocks,
)


class TestFinancialTools:
    """Test financial data tools."""
    
    def test_get_stock_quote(self):
        """Test getting stock quote."""
        result = get_stock_quote("AAPL")
        assert "symbol" in result
        assert result["symbol"] == "AAPL"
        assert "price" in result
        assert result.get("price") is not None
    
    def test_get_company_info(self):
        """Test getting company info."""
        result = get_company_info("MSFT")
        assert "symbol" in result
        assert result["symbol"] == "MSFT"
        assert "name" in result
    
    def test_get_stock_history(self):
        """Test getting stock history."""
        result = get_stock_history("GOOGL", period="1mo")
        assert "symbol" in result
        assert "history" in result
        assert len(result.get("history", [])) > 0
    
    def test_get_analyst_recommendations(self):
        """Test getting analyst recommendations."""
        result = get_analyst_recommendations("NVDA")
        assert "symbol" in result
        assert result["symbol"] == "NVDA"
    
    def test_technical_analysis(self):
        """Test technical analysis."""
        result = technical_analysis("TSLA")
        assert "symbol" in result
        assert "sma_20" in result
        assert "rsi" in result
        assert "signals" in result
    
    def test_compare_stocks(self):
        """Test comparing stocks."""
        result = compare_stocks(["AAPL", "MSFT", "GOOGL"])
        assert "comparison" in result
        assert len(result["comparison"]) == 3



"""Sigma - Elite Financial Research Agent."""

__version__ = "2.0.2"
__author__ = "Sigma"

from sigma.core.agent import SigmaAgent
from sigma.core.config import LLMProvider, get_settings

__all__ = ["SigmaAgent", "LLMProvider", "get_settings"]



"""Financial tools for real market data."""

import asyncio
from datetime import datetime, timedelta
from typing import Any, Optional, Union
import time

import httpx
import pandas as pd


# Tool registry
_tools: dict[str, dict[str, Any]] = {}


def tool(name: str, description: str, parameters: dict[str, Any]):
    """Register a tool."""
    def decorator(func):
        _tools[name] = {
            "name": name,
            "description": description,
            "parameters": parameters,
            "function": func,
        }
        return func
    return decorator


def get_all_tools() -> list[dict[str, Any]]:
    """Get all tool definitions."""
    return [
        {"name": t["name"], "description": t["description"], "parameters": t["parameters"]}
        for t in _tools.values()
    ]


async def execute_tool(name: str, arguments: dict[str, Any]) -> Any:
    """Execute a tool."""
    if name not in _tools:
        return {"error": f"Tool not found: {name}"}
    
    try:
        func = _tools[name]["function"]
        if asyncio.iscoroutinefunction(func):
            return await func(**arguments)
        return func(**arguments)
    except Exception as e:
        return {"error": str(e)}


def _format_date(val: Any) -> str:
    """Safely format a date value to string."""
    if hasattr(val, 'strftime'):
        return val.strftime("%Y-%m-%d")
    elif isinstance(val, pd.Timestamp):
        return val.strftime("%Y-%m-%d")
    return str(val)


def _is_dataframe(obj: Any) -> bool:
    """Check if object is a DataFrame."""
    return isinstance(obj, pd.DataFrame)


# ============================================================================
# MARKET DATA TOOLS (yfinance)
# ============================================================================

@tool(
    name="get_stock_quote",
    description="Get real-time stock quote with price, volume, market cap, P/E ratio, and other key metrics",
    parameters={
        "type": "object",
        "properties": {
            "symbol": {"type": "string", "description": "Stock ticker symbol (e.g., AAPL, GOOGL)"}
        },
        "required": ["symbol"]
    }
)
def get_stock_quote(symbol: str) -> dict[str, Any]:
    """Get stock quote using yfinance."""
    import yfinance as yf
    
    ticker = yf.Ticker(symbol.upper())
    info = ticker.info
    
    if not info or "regularMarketPrice" not in info:
        # Try fast_info
        try:
            fast = ticker.fast_info
            return {
                "symbol": symbol.upper(),
                "price": fast.get("lastPrice", fast.get("regularMarketPrice")),
                "previous_close": fast.get("previousClose"),
                "market_cap": fast.get("marketCap"),
                "error": None
            }
        except:
            return {"error": f"Could not fetch quote for {symbol}"}
    
    return {
        "symbol": symbol.upper(),
        "name": info.get("shortName", info.get("longName", "")),
        "price": info.get("regularMarketPrice", info.get("currentPrice")),
        "previous_close": info.get("previousClose"),
        "open": info.get("regularMarketOpen"),
        "day_high": info.get("regularMarketDayHigh"),
        "day_low": info.get("regularMarketDayLow"),
        "volume": info.get("regularMarketVolume"),
        "avg_volume": info.get("averageVolume"),
        "market_cap": info.get("marketCap"),
        "pe_ratio": info.get("trailingPE"),
        "forward_pe": info.get("forwardPE"),
        "eps": info.get("trailingEps"),
        "dividend_yield": info.get("dividendYield"),
        "52_week_high": info.get("fiftyTwoWeekHigh"),
        "52_week_low": info.get("fiftyTwoWeekLow"),
        "50_day_avg": info.get("fiftyDayAverage"),
        "200_day_avg": info.get("twoHundredDayAverage"),
        "beta": info.get("beta"),
        "currency": info.get("currency", "USD"),
    }


@tool(
    name="get_stock_history",
    description="Get historical price data for a stock",
    parameters={
        "type": "object",
        "properties": {
            "symbol": {"type": "string", "description": "Stock ticker symbol"},
            "period": {"type": "string", "description": "Time period: 1d, 5d, 1mo, 3mo, 6mo, 1y, 2y, 5y, 10y, ytd, max", "default": "1mo"},
            "interval": {"type": "string", "description": "Data interval: 1m, 2m, 5m, 15m, 30m, 60m, 90m, 1h, 1d, 5d, 1wk, 1mo", "default": "1d"}
        },
        "required": ["symbol"]
    }
)
def get_stock_history(symbol: str, period: str = "1mo", interval: str = "1d") -> dict[str, Any]:
    """Get historical stock data."""
    import yfinance as yf
    
    ticker = yf.Ticker(symbol.upper())
    hist = ticker.history(period=period, interval=interval)
    
    if hist.empty:
        return {"error": f"No history for {symbol}"}
    
    records = []
    for date, row in hist.iterrows():
        records.append({
            "date": _format_date(date),
            "open": round(row["Open"], 2) if row["Open"] else None,
            "high": round(row["High"], 2) if row["High"] else None,
            "low": round(row["Low"], 2) if row["Low"] else None,
            "close": round(row["Close"], 2) if row["Close"] else None,
            "volume": int(row["Volume"]) if row["Volume"] else None,
        })
    
    # Calculate returns
    if len(records) > 1:
        start_price = records[0]["close"]
        end_price = records[-1]["close"]
        if start_price and end_price:
            total_return = ((end_price - start_price) / start_price) * 100
        else:
            total_return = None
    else:
        total_return = None
    
    return {
        "symbol": symbol.upper(),
        "period": period,
        "interval": interval,
        "data_points": len(records),
        "start_date": records[0]["date"] if records else None,
        "end_date": records[-1]["date"] if records else None,
        "total_return_pct": round(total_return, 2) if total_return else None,
        "history": records[-20:],  # Last 20 for brevity
    }


@tool(
    name="get_company_info",
    description="Get detailed company information including business description, sector, industry, and key statistics",
    parameters={
        "type": "object",
        "properties": {
            "symbol": {"type": "string", "description": "Stock ticker symbol"}
        },
        "required": ["symbol"]
    }
)
def get_company_info(symbol: str) -> dict[str, Any]:
    """Get company information."""
    import yfinance as yf
    
    ticker = yf.Ticker(symbol.upper())
    info = ticker.info
    
    if not info:
        return {"error": f"Could not fetch info for {symbol}"}
    
    return {
        "symbol": symbol.upper(),
        "name": info.get("shortName", info.get("longName", "")),
        "description": info.get("longBusinessSummary", ""),
        "sector": info.get("sector"),
        "industry": info.get("industry"),
        "website": info.get("website"),
        "country": info.get("country"),
        "employees": info.get("fullTimeEmployees"),
        "market_cap": info.get("marketCap"),
        "enterprise_value": info.get("enterpriseValue"),
        "revenue": info.get("totalRevenue"),
        "gross_profit": info.get("grossProfits"),
        "ebitda": info.get("ebitda"),
        "net_income": info.get("netIncomeToCommon"),
        "profit_margin": info.get("profitMargins"),
        "operating_margin": info.get("operatingMargins"),
        "roe": info.get("returnOnEquity"),
        "roa": info.get("returnOnAssets"),
        "debt_to_equity": info.get("debtToEquity"),
        "current_ratio": info.get("currentRatio"),
        "quick_ratio": info.get("quickRatio"),
        "free_cash_flow": info.get("freeCashflow"),
    }


@tool(
    name="get_financial_statements",
    description="Get income statement, balance sheet, and cash flow statement data",
    parameters={
        "type": "object",
        "properties": {
            "symbol": {"type": "string", "description": "Stock ticker symbol"},
            "statement": {"type": "string", "description": "Statement type: income, balance, cashflow", "default": "income"},
            "period": {"type": "string", "description": "Period: annual or quarterly", "default": "annual"}
        },
        "required": ["symbol"]
    }
)
def get_financial_statements(symbol: str, statement: str = "income", period: str = "annual") -> dict[str, Any]:
    """Get financial statements."""
    import yfinance as yf
    
    ticker = yf.Ticker(symbol.upper())
    
    if statement == "income":
        df = ticker.income_stmt if period == "annual" else ticker.quarterly_income_stmt
    elif statement == "balance":
        df = ticker.balance_sheet if period == "annual" else ticker.quarterly_balance_sheet
    elif statement == "cashflow":
        df = ticker.cashflow if period == "annual" else ticker.quarterly_cashflow
    else:
        return {"error": f"Invalid statement type: {statement}"}
    
    if df is None or df.empty:
        return {"error": f"No {statement} statement for {symbol}"}
    
    # Convert to dict
    data = {}
    for col in df.columns[:4]:  # Last 4 periods
        period_key = _format_date(col)
        data[period_key] = {}
        for idx, val in df[col].items():
            if val is not None and not (isinstance(val, float) and val != val):  # Check for NaN
                data[period_key][str(idx)] = float(val) if isinstance(val, (int, float)) else val
    
    return {
        "symbol": symbol.upper(),
        "statement_type": statement,
        "period": period,
        "data": data,
    }


@tool(
    name="get_analyst_recommendations",
    description="Get analyst recommendations and price targets",
    parameters={
        "type": "object",
        "properties": {
            "symbol": {"type": "string", "description": "Stock ticker symbol"}
        },
        "required": ["symbol"]
    }
)
def get_analyst_recommendations(symbol: str) -> dict[str, Any]:
    """Get analyst recommendations."""
    import yfinance as yf
    
    ticker = yf.Ticker(symbol.upper())
    
    # Get recommendations
    recs = ticker.recommendations
    info = ticker.info
    
    result: dict[str, Any] = {
        "symbol": symbol.upper(),
        "recommendation": info.get("recommendationKey"),
        "mean_rating": info.get("recommendationMean"),
        "num_analysts": info.get("numberOfAnalystOpinions"),
        "target_high": info.get("targetHighPrice"),
        "target_low": info.get("targetLowPrice"),
        "target_mean": info.get("targetMeanPrice"),
        "target_median": info.get("targetMedianPrice"),
        "current_price": info.get("regularMarketPrice"),
    }
    
    # Add upside/downside
    if result.get("target_mean") and result.get("current_price"):
        upside = ((result["target_mean"] - result["current_price"]) / result["current_price"]) * 100
        result["upside_pct"] = round(upside, 2)
    
    # Recent recommendations
    if recs is not None and _is_dataframe(recs):
        recs_df: pd.DataFrame = recs  # type: ignore
        if not recs_df.empty:
            recent = recs_df.tail(10).to_dict("records")
            result["recent_recommendations"] = recent
    
    return result


@tool(
    name="get_insider_trades",
    description="Get recent insider trading activity",
    parameters={
        "type": "object",
        "properties": {
            "symbol": {"type": "string", "description": "Stock ticker symbol"}
        },
        "required": ["symbol"]
    }
)
def get_insider_trades(symbol: str) -> dict[str, Any]:
    """Get insider trades."""
    import yfinance as yf
    
    ticker = yf.Ticker(symbol.upper())
    insiders = ticker.insider_transactions
    
    if insiders is None or insiders.empty:
        return {"symbol": symbol.upper(), "trades": [], "message": "No insider trades found"}
    
    trades = []
    for _, row in insiders.head(20).iterrows():
        trade = {}
        for col in insiders.columns:
            val = row[col]
            if hasattr(val, 'strftime'):
                trade[col] = val.strftime("%Y-%m-%d")
            elif val is None or (isinstance(val, float) and val != val):
                trade[col] = None
            else:
                trade[col] = val
        trades.append(trade)
    
    return {
        "symbol": symbol.upper(),
        "total_trades": len(insiders),
        "trades": trades,
    }


@tool(
    name="get_institutional_holders",
    description="Get institutional ownership data",
    parameters={
        "type": "object",
        "properties": {
            "symbol": {"type": "string", "description": "Stock ticker symbol"}
        },
        "required": ["symbol"]
    }
)
def get_institutional_holders(symbol: str) -> dict[str, Any]:
    """Get institutional holders."""
    import yfinance as yf
    
    ticker = yf.Ticker(symbol.upper())
    holders = ticker.institutional_holders
    
    if holders is None or holders.empty:
        return {"symbol": symbol.upper(), "holders": [], "message": "No institutional holders found"}
    
    holder_list = []
    for _, row in holders.iterrows():
        holder = {}
        for col in holders.columns:
            val = row[col]
            if hasattr(val, 'strftime'):
                holder[col] = val.strftime("%Y-%m-%d")
            elif val is None or (isinstance(val, float) and val != val):
                holder[col] = None
            else:
                holder[col] = val
        holder_list.append(holder)
    
    return {
        "symbol": symbol.upper(),
        "holders": holder_list,
    }


@tool(
    name="get_earnings_calendar",
    description="Get earnings history and upcoming earnings dates",
    parameters={
        "type": "object",
        "properties": {
            "symbol": {"type": "string", "description": "Stock ticker symbol"}
        },
        "required": ["symbol"]
    }
)
def get_earnings_calendar(symbol: str) -> dict[str, Any]:
    """Get earnings data."""
    import yfinance as yf
    
    ticker = yf.Ticker(symbol.upper())
    earnings = ticker.earnings_history
    calendar = ticker.calendar
    
    result: dict[str, Any] = {"symbol": symbol.upper()}
    
    if earnings is not None and _is_dataframe(earnings) and not earnings.empty:
        history = []
        for _, row in earnings.iterrows():
            entry = {}
            for col in earnings.columns:
                val = row[col]
                if hasattr(val, 'strftime'):
                    entry[col] = _format_date(val)
                elif val is None or (isinstance(val, float) and val != val):
                    entry[col] = None
                else:
                    entry[col] = val
            history.append(entry)
        result["earnings_history"] = history
    
    if calendar is not None:
        try:
            if isinstance(calendar, dict):
                result["calendar"] = calendar
            elif hasattr(calendar, 'to_dict'):
                # DataFrame calendar
                raw_dict = dict(calendar.to_dict())  # type: ignore[union-attr]
                result["calendar"] = {str(k): str(v) if v is not None else None for k, v in raw_dict.items()}
            else:
                result["calendar"] = str(calendar)
        except Exception:
            result["calendar"] = str(calendar)
    
    return result


@tool(
    name="get_options_chain",
    description="Get options chain data including calls and puts",
    parameters={
        "type": "object",
        "properties": {
            "symbol": {"type": "string", "description": "Stock ticker symbol"},
            "expiration": {"type": "string", "description": "Expiration date (YYYY-MM-DD) or 'next' for nearest expiration"}
        },
        "required": ["symbol"]
    }
)
def get_options_chain(symbol: str, expiration: Optional[str] = None) -> dict[str, Any]:
    """Get options chain."""
    import yfinance as yf
    
    ticker = yf.Ticker(symbol.upper())
    expirations = ticker.options
    
    if not expirations:
        return {"error": f"No options available for {symbol}"}
    
    # Use first expiration if not specified
    exp_date = expiration if expiration and expiration != "next" else expirations[0]
    
    if exp_date not in expirations:
        return {"error": f"Invalid expiration. Available: {expirations[:5]}"}
    
    chain = ticker.option_chain(exp_date)
    
    def process_options(df, limit=10):
        options = []
        for _, row in df.head(limit).iterrows():
            opt = {}
            for col in df.columns:
                val = row[col]
                if val is None or (isinstance(val, float) and val != val):
                    opt[col] = None
                elif isinstance(val, (int, float)):
                    opt[col] = round(float(val), 4)
                else:
                    opt[col] = val
            options.append(opt)
        return options
    
    return {
        "symbol": symbol.upper(),
        "expiration": exp_date,
        "available_expirations": list(expirations[:10]),
        "calls": process_options(chain.calls),
        "puts": process_options(chain.puts),
    }


@tool(
    name="get_dividends",
    description="Get dividend history and yield information",
    parameters={
        "type": "object",
        "properties": {
            "symbol": {"type": "string", "description": "Stock ticker symbol"}
        },
        "required": ["symbol"]
    }
)
def get_dividends(symbol: str) -> dict[str, Any]:
    """Get dividend data."""
    import yfinance as yf
    
    ticker = yf.Ticker(symbol.upper())
    dividends = ticker.dividends
    info = ticker.info
    
    result: dict[str, Any] = {
        "symbol": symbol.upper(),
        "dividend_yield": info.get("dividendYield"),
        "dividend_rate": info.get("dividendRate"),
        "payout_ratio": info.get("payoutRatio"),
        "ex_dividend_date": info.get("exDividendDate"),
    }
    
    if dividends is not None and not dividends.empty:
        history = []
        for date, amount in dividends.tail(20).items():
            history.append({
                "date": _format_date(date),
                "amount": round(amount, 4)
            })
        result["history"] = history
        result["total_dividends_1y"] = round(dividends.tail(4).sum(), 4)
    
    return result


@tool(
    name="compare_stocks",
    description="Compare multiple stocks on key metrics",
    parameters={
        "type": "object",
        "properties": {
            "symbols": {"type": "array", "items": {"type": "string"}, "description": "List of stock symbols to compare"}
        },
        "required": ["symbols"]
    }
)
def compare_stocks(symbols: list[str]) -> dict[str, Any]:
    """Compare multiple stocks."""
    import yfinance as yf
    
    comparisons = []
    for symbol in symbols[:10]:  # Limit to 10
        ticker = yf.Ticker(symbol.upper())
        info = ticker.info
        
        if info:
            comparisons.append({
                "symbol": symbol.upper(),
                "name": info.get("shortName"),
                "price": info.get("regularMarketPrice"),
                "market_cap": info.get("marketCap"),
                "pe_ratio": info.get("trailingPE"),
                "forward_pe": info.get("forwardPE"),
                "peg_ratio": info.get("pegRatio"),
                "ps_ratio": info.get("priceToSalesTrailing12Months"),
                "pb_ratio": info.get("priceToBook"),
                "dividend_yield": info.get("dividendYield"),
                "profit_margin": info.get("profitMargins"),
                "roe": info.get("returnOnEquity"),
                "debt_to_equity": info.get("debtToEquity"),
                "revenue_growth": info.get("revenueGrowth"),
                "earnings_growth": info.get("earningsGrowth"),
                "52w_change": info.get("52WeekChange"),
            })
    
    return {
        "comparison": comparisons,
        "symbols_compared": len(comparisons),
    }


@tool(
    name="get_market_movers",
    description="Get top market gainers and losers",
    parameters={
        "type": "object",
        "properties": {
            "category": {"type": "string", "description": "Category: gainers, losers, or active", "default": "gainers"}
        },
        "required": []
    }
)
async def get_market_movers(category: str = "gainers") -> dict[str, Any]:
    """Get market movers."""
    import yfinance as yf
    
    # Major indices and popular stocks to check
    symbols = [
        "AAPL", "MSFT", "GOOGL", "AMZN", "NVDA", "META", "TSLA", "BRK-B",
        "JPM", "JNJ", "V", "UNH", "HD", "PG", "MA", "DIS", "PYPL", "NFLX",
        "ADBE", "CRM", "INTC", "AMD", "CSCO", "PFE", "MRK", "ABT", "TMO",
        "NKE", "COST", "WMT", "XOM", "CVX", "BA", "CAT", "GS", "MS"
    ]
    
    movers = []
    for symbol in symbols:
        try:
            ticker = yf.Ticker(symbol)
            info = ticker.info
            if info and info.get("regularMarketPrice") and info.get("previousClose"):
                price = info["regularMarketPrice"]
                prev = info["previousClose"]
                change = price - prev
                change_pct = (change / prev) * 100
                movers.append({
                    "symbol": symbol,
                    "name": info.get("shortName"),
                    "price": round(price, 2),
                    "change": round(change, 2),
                    "change_pct": round(change_pct, 2),
                    "volume": info.get("regularMarketVolume"),
                })
        except:
            continue
    
    if category == "gainers":
        movers.sort(key=lambda x: x["change_pct"], reverse=True)
    elif category == "losers":
        movers.sort(key=lambda x: x["change_pct"])
    else:  # active
        movers.sort(key=lambda x: x.get("volume", 0) or 0, reverse=True)
    
    return {
        "category": category,
        "movers": movers[:15],
    }


@tool(
    name="get_sector_performance",
    description="Get sector ETF performance",
    parameters={
        "type": "object",
        "properties": {},
        "required": []
    }
)
def get_sector_performance() -> dict[str, Any]:
    """Get sector performance via ETFs."""
    import yfinance as yf
    
    sectors = {
        "XLK": "Technology",
        "XLF": "Financials",
        "XLV": "Healthcare",
        "XLE": "Energy",
        "XLI": "Industrials",
        "XLY": "Consumer Discretionary",
        "XLP": "Consumer Staples",
        "XLU": "Utilities",
        "XLB": "Materials",
        "XLRE": "Real Estate",
        "XLC": "Communication Services",
    }
    
    performance = []
    for symbol, name in sectors.items():
        try:
            ticker = yf.Ticker(symbol)
            info = ticker.info
            if info and info.get("regularMarketPrice") and info.get("previousClose"):
                price = info["regularMarketPrice"]
                prev = info["previousClose"]
                change_pct = ((price - prev) / prev) * 100
                performance.append({
                    "sector": name,
                    "etf": symbol,
                    "price": round(price, 2),
                    "change_pct": round(change_pct, 2),
                    "52w_change": info.get("52WeekChange"),
                })
        except:
            continue
    
    performance.sort(key=lambda x: x["change_pct"], reverse=True)
    
    return {
        "sectors": performance,
        "best_sector": performance[0]["sector"] if performance else None,
        "worst_sector": performance[-1]["sector"] if performance else None,
    }


@tool(
    name="get_market_indices",
    description="Get major market indices",
    parameters={
        "type": "object",
        "properties": {},
        "required": []
    }
)
def get_market_indices() -> dict[str, Any]:
    """Get market indices."""
    import yfinance as yf
    
    indices = {
        "^GSPC": "S&P 500",
        "^DJI": "Dow Jones",
        "^IXIC": "NASDAQ",
        "^RUT": "Russell 2000",
        "^VIX": "VIX",
        "^TNX": "10Y Treasury",
        "GC=F": "Gold",
        "CL=F": "Crude Oil",
        "BTC-USD": "Bitcoin",
    }
    
    data = []
    for symbol, name in indices.items():
        try:
            ticker = yf.Ticker(symbol)
            info = ticker.info
            if info:
                price = info.get("regularMarketPrice", info.get("previousClose"))
                prev = info.get("previousClose")
                if price and prev:
                    change_pct = ((price - prev) / prev) * 100
                else:
                    change_pct = None
                data.append({
                    "name": name,
                    "symbol": symbol,
                    "price": round(price, 2) if price else None,
                    "change_pct": round(change_pct, 2) if change_pct else None,
                })
        except:
            continue
    
    return {"indices": data}


@tool(
    name="calculate_portfolio_metrics",
    description="Calculate portfolio metrics given holdings",
    parameters={
        "type": "object",
        "properties": {
            "holdings": {
                "type": "array",
                "items": {
                    "type": "object",
                    "properties": {
                        "symbol": {"type": "string"},
                        "shares": {"type": "number"},
                        "cost_basis": {"type": "number"}
                    }
                },
                "description": "List of holdings with symbol, shares, and cost basis"
            }
        },
        "required": ["holdings"]
    }
)
def calculate_portfolio_metrics(holdings: list[dict]) -> dict[str, Any]:
    """Calculate portfolio metrics."""
    import yfinance as yf
    
    results = []
    total_value = 0
    total_cost = 0
    
    for holding in holdings:
        symbol = holding["symbol"]
        shares = holding["shares"]
        cost_basis = holding.get("cost_basis", 0)
        
        ticker = yf.Ticker(symbol.upper())
        info = ticker.info
        
        if info and info.get("regularMarketPrice"):
            price = info["regularMarketPrice"]
            value = price * shares
            cost = cost_basis * shares if cost_basis else 0
            gain = value - cost if cost else None
            gain_pct = ((value - cost) / cost * 100) if cost else None
            
            results.append({
                "symbol": symbol.upper(),
                "shares": shares,
                "price": round(price, 2),
                "value": round(value, 2),
                "cost_basis": cost_basis,
                "total_cost": round(cost, 2),
                "gain": round(gain, 2) if gain else None,
                "gain_pct": round(gain_pct, 2) if gain_pct else None,
            })
            
            total_value += value
            total_cost += cost
    
    total_gain = total_value - total_cost if total_cost else None
    total_gain_pct = ((total_value - total_cost) / total_cost * 100) if total_cost else None
    
    # Calculate weights
    for r in results:
        r["weight_pct"] = round((r["value"] / total_value * 100), 2) if total_value else 0
    
    return {
        "holdings": results,
        "total_value": round(total_value, 2),
        "total_cost": round(total_cost, 2),
        "total_gain": round(total_gain, 2) if total_gain else None,
        "total_gain_pct": round(total_gain_pct, 2) if total_gain_pct else None,
    }


@tool(
    name="search_stocks",
    description="Search for stocks by company name or symbol",
    parameters={
        "type": "object",
        "properties": {
            "query": {"type": "string", "description": "Search query (company name or symbol)"}
        },
        "required": ["query"]
    }
)
async def search_stocks(query: str) -> dict[str, Any]:
    """Search for stocks."""
    url = f"https://query1.finance.yahoo.com/v1/finance/search?q={query}&quotesCount=10&newsCount=0"
    
    async with httpx.AsyncClient() as client:
        resp = await client.get(url, headers={"User-Agent": "Mozilla/5.0"})
        data = resp.json()
    
    results = []
    for quote in data.get("quotes", []):
        if quote.get("quoteType") == "EQUITY":
            results.append({
                "symbol": quote.get("symbol"),
                "name": quote.get("shortname") or quote.get("longname"),
                "exchange": quote.get("exchange"),
                "type": quote.get("typeDisp"),
            })
    
    return {"results": results, "query": query}


@tool(
    name="get_stock_news",
    description="Get recent news for a stock",
    parameters={
        "type": "object",
        "properties": {
            "symbol": {"type": "string", "description": "Stock ticker symbol"}
        },
        "required": ["symbol"]
    }
)
def get_stock_news(symbol: str) -> dict[str, Any]:
    """Get stock news."""
    import yfinance as yf
    
    ticker = yf.Ticker(symbol.upper())
    news = ticker.news
    
    if not news:
        return {"symbol": symbol.upper(), "news": [], "message": "No news found"}
    
    articles = []
    for item in news[:10]:
        articles.append({
            "title": item.get("title"),
            "publisher": item.get("publisher"),
            "link": item.get("link"),
            "published": datetime.fromtimestamp(item.get("providerPublishTime", 0)).isoformat() if item.get("providerPublishTime") else None,
            "type": item.get("type"),
        })
    
    return {
        "symbol": symbol.upper(),
        "news": articles,
    }


@tool(
    name="technical_analysis",
    description="Perform basic technical analysis on a stock",
    parameters={
        "type": "object",
        "properties": {
            "symbol": {"type": "string", "description": "Stock ticker symbol"}
        },
        "required": ["symbol"]
    }
)
def technical_analysis(symbol: str) -> dict[str, Any]:
    """Perform technical analysis."""
    import yfinance as yf
    import numpy as np
    
    ticker = yf.Ticker(symbol.upper())
    hist = ticker.history(period="6mo")
    
    if hist.empty:
        return {"error": f"No data for {symbol}"}
    
    close = hist["Close"]
    
    # Moving averages
    sma_20 = close.rolling(window=20).mean().iloc[-1]
    sma_50 = close.rolling(window=50).mean().iloc[-1]
    sma_200 = close.rolling(window=200).mean().iloc[-1] if len(close) >= 200 else None
    
    # RSI
    delta = close.diff()
    gain = (delta.where(delta > 0, 0)).rolling(window=14).mean()
    loss = (-delta.where(delta < 0, 0)).rolling(window=14).mean()
    rs = gain / loss
    rsi = 100 - (100 / (1 + rs)).iloc[-1]
    
    # MACD
    ema_12 = close.ewm(span=12, adjust=False).mean()
    ema_26 = close.ewm(span=26, adjust=False).mean()
    macd = ema_12 - ema_26
    signal = macd.ewm(span=9, adjust=False).mean()
    
    current_price = close.iloc[-1]
    
    # Signals
    signals = []
    if current_price > sma_20:
        signals.append("Above SMA20 (bullish)")
    else:
        signals.append("Below SMA20 (bearish)")
    
    if current_price > sma_50:
        signals.append("Above SMA50 (bullish)")
    else:
        signals.append("Below SMA50 (bearish)")
    
    if rsi > 70:
        signals.append("RSI > 70 (overbought)")
    elif rsi < 30:
        signals.append("RSI < 30 (oversold)")
    else:
        signals.append(f"RSI neutral ({round(rsi, 1)})")
    
    if macd.iloc[-1] > signal.iloc[-1]:
        signals.append("MACD above signal (bullish)")
    else:
        signals.append("MACD below signal (bearish)")
    
    return {
        "symbol": symbol.upper(),
        "price": round(current_price, 2),
        "sma_20": round(sma_20, 2),
        "sma_50": round(sma_50, 2),
        "sma_200": round(sma_200, 2) if sma_200 else None,
        "rsi": round(rsi, 2),
        "macd": round(macd.iloc[-1], 4),
        "macd_signal": round(signal.iloc[-1], 4),
        "signals": signals,
        "support": round(hist["Low"].tail(20).min(), 2),
        "resistance": round(hist["High"].tail(20).max(), 2),
    }


# ============================================================================
# CHART TOOLS
# ============================================================================

@tool(
    name="generate_price_chart",
    description="Generate a beautiful ASCII price chart in the terminal for a stock",
    parameters={
        "type": "object",
        "properties": {
            "symbol": {"type": "string", "description": "Stock ticker symbol"},
            "period": {"type": "string", "description": "Time period: 1mo, 3mo, 6mo, 1y, 2y", "default": "3mo"},
            "chart_type": {"type": "string", "description": "Chart type: line, candle, bar, area", "default": "line"},
            "show_volume": {"type": "boolean", "description": "Show volume subplot", "default": True}
        },
        "required": ["symbol"]
    }
)
def generate_price_chart(symbol: str, period: str = "3mo", chart_type: str = "line", show_volume: bool = True) -> dict[str, Any]:
    """Generate a price chart."""
    from sigma.tools.charts import create_price_chart
    
    chart = create_price_chart(symbol, period, chart_type, show_volume)
    return {
        "symbol": symbol.upper(),
        "period": period,
        "chart_type": chart_type,
        "chart": chart,
        "display_as_chart": True,  # Flag for UI to render properly
    }


@tool(
    name="generate_comparison_chart",
    description="Generate a comparison chart for multiple stocks showing percentage change",
    parameters={
        "type": "object",
        "properties": {
            "symbols": {"type": "array", "items": {"type": "string"}, "description": "List of stock symbols to compare"},
            "period": {"type": "string", "description": "Time period", "default": "3mo"}
        },
        "required": ["symbols"]
    }
)
def generate_comparison_chart(symbols: list[str], period: str = "3mo") -> dict[str, Any]:
    """Generate a comparison chart."""
    from sigma.tools.charts import create_comparison_chart
    
    chart = create_comparison_chart(symbols, period)
    return {
        "symbols": [s.upper() for s in symbols],
        "period": period,
        "chart": chart,
        "display_as_chart": True,
    }


@tool(
    name="generate_rsi_chart",
    description="Generate a price chart with RSI indicator showing overbought/oversold levels",
    parameters={
        "type": "object",
        "properties": {
            "symbol": {"type": "string", "description": "Stock ticker symbol"},
            "period": {"type": "string", "description": "Time period", "default": "3mo"}
        },
        "required": ["symbol"]
    }
)
def generate_rsi_chart(symbol: str, period: str = "3mo") -> dict[str, Any]:
    """Generate RSI chart."""
    from sigma.tools.charts import create_rsi_chart
    
    chart = create_rsi_chart(symbol, period)
    return {
        "symbol": symbol.upper(),
        "period": period,
        "chart": chart,
        "display_as_chart": True,
    }


@tool(
    name="generate_sector_chart",
    description="Generate a sector performance chart showing daily changes for all major sectors",
    parameters={
        "type": "object",
        "properties": {},
        "required": []
    }
)
def generate_sector_chart() -> dict[str, Any]:
    """Generate sector chart."""
    from sigma.tools.charts import create_sector_chart
    
    chart = create_sector_chart()
    return {
        "chart": chart,
        "display_as_chart": True,
    }


# ============================================================================
# BACKTEST TOOLS
# ============================================================================

@tool(
    name="list_backtest_strategies",
    description="List all available backtesting strategies with descriptions",
    parameters={
        "type": "object",
        "properties": {},
        "required": []
    }
)
def list_backtest_strategies() -> dict[str, Any]:
    """List available strategies."""
    from sigma.tools.backtest import get_available_strategies
    
    strategies = get_available_strategies()
    return {
        "strategies": strategies,
        "count": len(strategies),
    }


@tool(
    name="generate_backtest",
    description="Generate a LEAN engine compatible backtest algorithm for a trading strategy",
    parameters={
        "type": "object",
        "properties": {
            "symbol": {"type": "string", "description": "Stock ticker symbol"},
            "strategy": {"type": "string", "description": "Strategy type: sma_crossover, rsi_mean_reversion, macd_momentum, bollinger_bands, dual_momentum, breakout"},
            "start_date": {"type": "string", "description": "Start date (YYYY-MM-DD)", "default": None},
            "end_date": {"type": "string", "description": "End date (YYYY-MM-DD)", "default": None},
            "initial_capital": {"type": "number", "description": "Starting capital", "default": 100000},
            "params": {"type": "object", "description": "Strategy-specific parameters (optional)"}
        },
        "required": ["symbol", "strategy"]
    }
)
def generate_backtest(
    symbol: str,
    strategy: str,
    start_date: Optional[str] = None,
    end_date: Optional[str] = None,
    initial_capital: float = 100000,
    params: Optional[dict] = None,
) -> dict[str, Any]:
    """Generate a backtest algorithm."""
    from sigma.tools.backtest import generate_lean_algorithm
    
    result = generate_lean_algorithm(
        symbol=symbol,
        strategy=strategy,
        start_date=start_date,
        end_date=end_date,
        initial_capital=initial_capital,
        params=params,
    )
    return result


@tool(
    name="generate_custom_backtest",
    description="Generate a custom backtest algorithm with user-specified entry/exit conditions",
    parameters={
        "type": "object",
        "properties": {
            "symbol": {"type": "string", "description": "Stock ticker symbol"},
            "entry_conditions": {"type": "array", "items": {"type": "string"}, "description": "List of entry conditions (e.g., 'RSI below 30', 'Price above SMA 50')"},
            "exit_conditions": {"type": "array", "items": {"type": "string"}, "description": "List of exit conditions"},
            "indicators": {"type": "array", "items": {"type": "string"}, "description": "Indicators to use: sma, ema, rsi, macd, bb, atr, adx"},
            "initial_capital": {"type": "number", "description": "Starting capital", "default": 100000}
        },
        "required": ["symbol", "entry_conditions", "exit_conditions", "indicators"]
    }
)
def generate_custom_backtest(
    symbol: str,
    entry_conditions: list[str],
    exit_conditions: list[str],
    indicators: list[str],
    initial_capital: float = 100000,
) -> dict[str, Any]:
    """Generate custom backtest."""
    from sigma.tools.backtest import generate_custom_algorithm
    
    return generate_custom_algorithm(
        symbol=symbol,
        entry_conditions=entry_conditions,
        exit_conditions=exit_conditions,
        indicators=indicators,
        initial_capital=initial_capital,
    )


@tool(
    name="run_backtest",
    description="Run a full backtest using LEAN CLI. Installs LEAN if needed, generates algorithm, and runs backtest.",
    parameters={
        "type": "object",
        "properties": {
            "symbol": {"type": "string", "description": "Stock ticker symbol"},
            "strategy": {
                "type": "string",
                "description": "Strategy: sma_crossover, rsi_mean_reversion, macd_momentum, bollinger_bands, dual_momentum, breakout",
                "default": "sma_crossover"
            },
            "initial_capital": {"type": "number", "description": "Starting capital", "default": 100000}
        },
        "required": ["symbol"]
    }
)
def run_backtest_tool(
    symbol: str,
    strategy: str = "sma_crossover",
    initial_capital: float = 100000,
) -> dict[str, Any]:
    """Run a full backtest with LEAN."""
    from sigma.tools.backtest import run_backtest
    
    return run_backtest(
        symbol=symbol,
        strategy=strategy,
        initial_capital=initial_capital,
    )


@tool(
    name="check_lean_status",
    description="Check if LEAN CLI is installed and provide setup instructions",
    parameters={
        "type": "object",
        "properties": {},
        "required": []
    }
)
def check_lean_status_tool() -> dict[str, Any]:
    """Check LEAN installation status."""
    from sigma.tools.backtest import check_lean_status
    
    return check_lean_status()


# ============================================================================
# PREDICTION TOOLS
# ============================================================================

@tool(
    name="price_forecast",
    description="Generate price predictions using multiple technical models",
    parameters={
        "type": "object",
        "properties": {
            "symbol": {"type": "string", "description": "Stock ticker symbol"},
            "horizon": {"type": "string", "description": "Forecast horizon: 1w, 1m, 3m, 6m", "default": "1m"}
        },
        "required": ["symbol"]
    }
)
def price_forecast(symbol: str, horizon: str = "1m") -> dict[str, Any]:
    """Generate price predictions."""
    import yfinance as yf
    import numpy as np
    
    ticker = yf.Ticker(symbol.upper())
    hist = ticker.history(period="1y")
    
    if hist.empty:
        return {"error": f"No data for {symbol}"}
    
    # Convert to numpy array explicitly
    close = np.array(hist["Close"].values, dtype=float)
    current_price = float(close[-1])
    
    # Map horizon to days
    horizon_days = {"1w": 5, "1m": 21, "3m": 63, "6m": 126}.get(horizon, 21)
    
    # Multiple prediction methods
    predictions = {}
    
    # 1. Simple Moving Average Projection
    sma_20 = float(np.mean(close[-20:]))
    sma_50 = float(np.mean(close[-50:])) if len(close) >= 50 else sma_20
    sma_trend = (sma_20 - sma_50) / sma_50 if sma_50 else 0
    sma_prediction = current_price * (1 + sma_trend * (horizon_days / 20))
    predictions["sma_trend"] = round(sma_prediction, 2)
    
    # 2. Linear Regression
    x = np.arange(len(close))
    coeffs = np.polyfit(x, close, 1)
    lr_prediction = float(coeffs[0]) * (len(close) + horizon_days) + float(coeffs[1])
    predictions["linear_regression"] = round(lr_prediction, 2)
    
    # 3. Volatility-adjusted (Monte Carlo simple)
    daily_returns = np.diff(close) / close[:-1]
    avg_return = float(np.mean(daily_returns))
    std_return = float(np.std(daily_returns))
    
    # Expected value with drift
    mc_prediction = current_price * (1 + avg_return * horizon_days)
    predictions["monte_carlo_expected"] = round(mc_prediction, 2)
    
    # Confidence interval (1 std)
    upper = current_price * np.exp((avg_return - 0.5 * std_return**2) * horizon_days + std_return * np.sqrt(horizon_days) * 1.96)
    lower = current_price * np.exp((avg_return - 0.5 * std_return**2) * horizon_days - std_return * np.sqrt(horizon_days) * 1.96)
    
    # 4. Mean reversion to SMA 200
    sma_200 = float(np.mean(close[-200:])) if len(close) >= 200 else sma_50
    mean_rev_prediction = current_price + (sma_200 - current_price) * 0.3  # 30% reversion
    predictions["mean_reversion"] = round(mean_rev_prediction, 2)
    
    # Consensus
    consensus = float(np.mean(list(predictions.values())))
    
    # Analyst targets for comparison
    info = ticker.info
    analyst_target = info.get("targetMeanPrice")
    
    return {
        "symbol": symbol.upper(),
        "current_price": round(current_price, 2),
        "horizon": horizon,
        "horizon_days": horizon_days,
        "predictions": predictions,
        "consensus_prediction": round(consensus, 2),
        "confidence_interval": {
            "lower_95": round(lower, 2),
            "upper_95": round(upper, 2),
        },
        "analyst_target": analyst_target,
        "volatility": {
            "daily": round(std_return * 100, 2),
            "annualized": round(std_return * np.sqrt(252) * 100, 2),
        },
        "disclaimer": "Predictions are based on historical patterns and should not be considered financial advice."
    }


@tool(
    name="sentiment_analysis",
    description="Analyze market sentiment for a stock based on various signals",
    parameters={
        "type": "object",
        "properties": {
            "symbol": {"type": "string", "description": "Stock ticker symbol"}
        },
        "required": ["symbol"]
    }
)
def sentiment_analysis(symbol: str) -> dict[str, Any]:
    """Analyze sentiment from multiple sources."""
    import yfinance as yf
    
    ticker = yf.Ticker(symbol.upper())
    info = ticker.info
    hist = ticker.history(period="3mo")
    
    if hist.empty:
        return {"error": f"No data for {symbol}"}
    
    signals = []
    score = 0
    
    # 1. Analyst sentiment
    rec = info.get("recommendationKey", "").lower()
    if "strong buy" in rec or rec == "buy":
        signals.append(("Analyst Rating", rec.upper(), "bullish", 2))
        score += 2
    elif "hold" in rec:
        signals.append(("Analyst Rating", rec.upper(), "neutral", 0))
    elif "sell" in rec:
        signals.append(("Analyst Rating", rec.upper(), "bearish", -2))
        score -= 2
    
    # 2. Price vs 52-week range
    high_52 = info.get("fiftyTwoWeekHigh", 0)
    low_52 = info.get("fiftyTwoWeekLow", 0)
    price = info.get("regularMarketPrice", 0)
    
    if high_52 and low_52 and price:
        position = (price - low_52) / (high_52 - low_52) if high_52 != low_52 else 0.5
        if position > 0.8:
            signals.append(("52-Week Position", f"{position*100:.0f}%", "overbought", -1))
            score -= 1
        elif position < 0.2:
            signals.append(("52-Week Position", f"{position*100:.0f}%", "oversold", 1))
            score += 1
        else:
            signals.append(("52-Week Position", f"{position*100:.0f}%", "neutral", 0))
    
    # 3. RSI
    close = hist["Close"].values
    if len(close) > 14:
        deltas = [close[i] - close[i-1] for i in range(1, len(close))]
        gains = [d if d > 0 else 0 for d in deltas]
        losses = [-d if d < 0 else 0 for d in deltas]
        avg_gain = sum(gains[-14:]) / 14
        avg_loss = sum(losses[-14:]) / 14
        if avg_loss > 0:
            rs = avg_gain / avg_loss
            rsi = 100 - (100 / (1 + rs))
        else:
            rsi = 100
        
        if rsi > 70:
            signals.append(("RSI", f"{rsi:.1f}", "overbought", -1))
            score -= 1
        elif rsi < 30:
            signals.append(("RSI", f"{rsi:.1f}", "oversold", 1))
            score += 1
        else:
            signals.append(("RSI", f"{rsi:.1f}", "neutral", 0))
    
    # 4. Moving average trend
    sma_50 = sum(close[-50:]) / 50 if len(close) >= 50 else None
    sma_200 = sum(close[-200:]) / 200 if len(close) >= 200 else None
    
    if sma_50 and sma_200:
        if sma_50 > sma_200:
            signals.append(("Golden Cross", "SMA50 > SMA200", "bullish", 1))
            score += 1
        else:
            signals.append(("Death Cross", "SMA50 < SMA200", "bearish", -1))
            score -= 1
    
    # 5. Volume trend
    avg_vol = info.get("averageVolume", 0)
    curr_vol = info.get("regularMarketVolume", 0)
    
    if avg_vol and curr_vol:
        vol_ratio = curr_vol / avg_vol
        if vol_ratio > 1.5:
            signals.append(("Volume Surge", f"{vol_ratio:.1f}x avg", "high_interest", 0))
    
    # 6. Insider activity
    try:
        insiders = ticker.insider_transactions
        if insiders is not None and not insiders.empty:
            buys = insiders[insiders["Shares"].fillna(0) > 0].shape[0]
            sells = insiders[insiders["Shares"].fillna(0) < 0].shape[0]
            if buys > sells * 2:
                signals.append(("Insider Activity", f"{buys} buys vs {sells} sells", "bullish", 1))
                score += 1
            elif sells > buys * 2:
                signals.append(("Insider Activity", f"{sells} sells vs {buys} buys", "bearish", -1))
                score -= 1
    except:
        pass
    
    # Overall sentiment
    if score >= 3:
        overall = "STRONGLY BULLISH"
    elif score >= 1:
        overall = "BULLISH"
    elif score <= -3:
        overall = "STRONGLY BEARISH"
    elif score <= -1:
        overall = "BEARISH"
    else:
        overall = "NEUTRAL"
    
    return {
        "symbol": symbol.upper(),
        "overall_sentiment": overall,
        "sentiment_score": score,
        "max_score": 5,
        "min_score": -5,
        "signals": [{"indicator": s[0], "value": s[1], "signal": s[2], "score": s[3]} for s in signals],
    }



"""Core module initialization."""

from sigma.core.agent import SigmaAgent
from sigma.core.config import LLMProvider, Settings, get_settings
from sigma.core.llm import get_llm
from sigma.core.models import Message, MessageRole, ToolCall, ToolResult

__all__ = [
    "SigmaAgent",
    "LLMProvider",
    "Settings",
    "get_settings",
    "get_llm",
    "Message",
    "MessageRole",
    "ToolCall",
    "ToolResult",
]
